"""Tools for working with Q-grams, short sections of a string. For more
information, see this paper:

http://citeseerx.ist.psu.edu/viewdoc/summary?doi=10.1.1.14.6750

Note that we currently ignore the positions of the Q-grams, and simply compare
the set of Q-grams from each string.
"""
from collections import defaultdict

START="^"
END="$"

def qgrams(s, q=3):
    """Generate a set of Q-grams from a string."""
    s = START*(q-1) + s + END*(q-1)
    return frozenset(s[i:i+q] for i in range(len(s)-(q-1)))

def compare_strings(s1, s2, q=3):
    """Return the proportion of Q-grams that match between two strings, as a
    float between 0 and 1."""
    q1, q2 = qgrams(s1), qgrams(s2)
    return overlap(q1, q2)
    
def overlap(q1, q2):
    """Return the overlap of two sets, as a proportion of the size of the
    larger set."""
    return len(q1.intersection(q2))/max(len(q1), len(q2))

class QGramIndex:
    """An index to do fuzzy string matching.
    
    Strings are stored by the first idx_chars letters to speed lookups, so it
    will only find matches where these letters are the same. To do full fuzzy
    matching, create the index with idx_chars=0.
    """
    def __init__(self, q=3, idx_chars=3):
        self.q = q
        self.idx_chars = idx_chars
        self.data = defaultdict(set)

    def add(self, s):
        self.data[s[:self.idx_chars]].add((s, qgrams(s, q=self.q)))
    
    def discard(self, s):
        self.data[s[:self.idx_chars]].discard((s, qgrams(s, q=self.q)))
    
    def matches(self, s, threshold=0.85):
        """Find fuzzy matches for the given string. Yields 2-tuples of (string, 
        proportion matched), where proportion matched is above the given
        threshold."""
        target_qgrams = qgrams(s)
        for check_str, check_qgrams in self.data[s[:self.idx_chars]]:
            match_val = overlap(target_qgrams, check_qgrams)
            if match_val > threshold:
                yield check_str, match_val
    
    def best_match(self, s, threshold=0.85):
        """Return a 2-tuple of (string, proportion matched) for the best match
        to the given string. Raises KeyError if no matches are found for that
        threshold."""
        matches = sorted(self.matches(s, threshold=threshold), key=lambda x:x[1])
        if matches:
            return matches[-1]
        raise KeyError("No matches found for {0!r} at threshold {1}".format(s, threshold))



#! /usr/bin/env python
# -*- coding: utf8 -*-
"""
This modules is for loading and saving taxa in the JSON lines format. This is a
modification of JSON where a series of JSON items are stored line by lines,
allowing them to be decoded individually, without having to hold the entire
object in memory.
"""
import json

from taxonome import Taxon, TaxonSet

def iter_taxa(filehandle, progress=None):
    """Loads taxa from a JSON lines file, iterating over them, without
    keeping the entire set in memory."""
    i = 0
    for line in filehandle:
        d = json.loads(line)
        yield Taxon.from_dict(d)
        
        i += 1
        if progress and (i % 50 == 0):
            progress(filehandle.buffer.tell())

def load_taxa(filehandle, progress=None):
    """Load a TaxonSet from a JSON lines file."""
    ts = TaxonSet()
    for taxon in iter_taxa(filehandle, progress=progress):
        ts.add(taxon)
    return ts
    
def save_taxon(filehandle, taxon):
    """Saves a single taxon to the given file. Can be called repeatedly to save
    successive taxa."""
    s = json.dumps(taxon.to_dict(), indent=None)
    filehandle.write(s + "\n")

def save_taxa(filehandle, taxasource):
    """Saves a TaxonSet (or any iterable source of taxa) to the given handle, in
    JSON lines format."""
    for taxon in taxasource:
        save_taxon(filehandle, taxon)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module fetches information from the MBG Tropicos database.

http://www.tropicos.org/
"""
import json
from urllib.parse import urlencode
from urllib.request import urlopen

from taxonome import Name, Taxon
from taxonome.taxa.collection import TaxaResource
from taxonome.config import config

apikey = config['api-keys']['tropicos']

class TropicosError(KeyError): pass
class TropicosNoAPIKey(TropicosError):
    def __str__(self):
        return ("An API key is needed for Tropicos. Please visit "
                "http://services.tropicos.org/help?requestkey\n\n"
                "See the doc page on configuration for where to put the key.")

urlbase = "http://services.tropicos.org/Name/"
common_params = dict(apikey=apikey, format='json')

def _make_name(res):
    name = res["ScientificName"]
    auth = res["ScientificNameWithAuthors"].replace(name, "")
    return Name(name, auth)

def _load(url):
    res = json.loads(urlopen(url).read().decode())
    if "Error" in res[0]:
        raise TropicosError(res[0]["Error"])
    return res

def _search(name, wildcard=False):
    """Search for the specified name with the tropicos API"""
    if not common_params['apikey']:
        raise TropicosNoAPIKey
    params = {'name':name, 'type': 'wildcard' if wildcard else 'exact'}
    params.update(common_params)
    return _load(urlbase + "Search?" + urlencode(params))

def _summary(name_id):
    """Get summary info."""
    if not common_params['apikey']:
        raise TropicosNoAPIKey
    return _load(urlbase + name_id + "?" + urlencode(common_params))

def _call_method(method, name_id):
    """Get info for a given name_id. method is the name of one of the API methods,
    such as Synonyms or HigherTaxa. Details are on the Tropicos website:
    http://services.tropicos.org/help
    """
    if not common_params['apikey']:
        raise TropicosNoAPIKey
    url = urlbase + name_id + "/" + method + "?" + urlencode(common_params)
    return _load(url)

class TropicosTaxon(Taxon):
    id = ""
    @property
    def url(self):
        return "http://www.tropicos.org/Name/" + id

class TropicosTaxaResource(TaxaResource):
    def resolve_name(self, key):
        plainkey = key.plain if hasattr(key, "plain") else key
        names = []
        for res in _search(plainkey):
            name = _make_name(res)
            name.id = str(res["NameId"])
            if name.match(key):
                try:
                    r2 = _call_method("AcceptedNames", name.id)[0]["AcceptedName"]
                    accname = _make_name(r2)
                    accname.id = str(r2["NameId"])
                except TropicosError:
                    accname = name
                names.append((name, accname))
        return names
    
    def get_by_accepted_name(self, name):
        if not hasattr(name, "id"):
            name.id = str(_search(name.name)["NameId"])
        t = TropicosTaxon(name)
        t.id = name.id
        return t

_instance = TropicosTaxaResource()

resolve_name = _instance.resolve_name
select = _instance.select
    



# coding: utf-8
"""
Module for scraping data from the USDA GRIN database.

Note that GRIN (and this module) works with a separate set of IDs from the PI
PI numbers used for accessions. These are on the end of URLs for pages. E.g.
the URL for accession PI 219824 is:
http://www.ars-grin.gov/cgi-bin/npgs/acc/display.pl?1178105
Hence the ID to use for the accession is 1178105.
"""

SEARCHb = "http://www.ars-grin.gov/cgi-bin/npgs/acc/search.pl?"
TAX_ACCb = "http://www.ars-grin.gov/cgi-bin/npgs/html/tax_acc.pl"
OBSb = "http://www.ars-grin.gov/cgi-bin/npgs/acc/obs.pl?"
ACCb = "http://www.ars-grin.gov/cgi-bin/npgs/acc/display.pl?"
TAX_SEARCH_URL = "http://www.ars-grin.gov/cgi-bin/npgs/html/tax_search.pl"
TAXON_URL = "http://www.ars-grin.gov/cgi-bin/npgs/html/taxon.pl?"
import re
from collections import defaultdict
from urllib.parse import urlencode
import warnings
from bs4 import BeautifulSoup

from taxonome import Name, Taxon, TaxonSet
from taxonome.taxa.base import UncertainSpeciesError
from taxonome.taxa.collection import TaxaResource
from .utils import urlopen

locnre = re.compile(r'Collected in: (.+?)<')
improvementre = re.compile(r'Improvement status: (.+?)\.')

improve_status = {'100':"Wild",
    '110':"Wild: Natural",
    '120':"Semi-natural/wild",
    '200':"Weedy",
    '300':"Landrace",
    '400':"Breeding/research material",
    '410':"Breeder's line",
    '411':"Synthetic population",
    '412':"Hybrid (breeding/research material)",
    '413':"Founder stock/base population",
    '414':"Inbred line (parent of hybrid cultivar)",
    '415':"Segregating population (breeding/research material)",
    '420':"Mutant/genetic stock",
    '500':"Advanced/improved cultivar",
    '999':"(Other)"}
improve_status_code = dict((v, k) for k, v in improve_status.items())

def _parse_name_item(tag):
    if isinstance(tag.contents[-1], str):
        auth = tag.contents[-1].strip()
        nameparts = tag.contents[:-1]
    else:
        auth = ""
        nameparts = tag.contents 
    rawname = "".join(a.text if hasattr(a, 'text') else a 
                    for a in nameparts).replace("&#215;", "×")
    if rawname.endswith((" spp.", " sp."," hybr.")):
        raise UncertainSpeciesError
    return Name(rawname, auth)

def _parse_synonym(tag):
    accname = _parse_name_item(tag.b)
    contents = tag.contents[:-2]
    auth = contents[-1].rstrip("(=").strip()
    rawname = "".join(a.text if hasattr(a, 'text') else a 
                    for a in contents[:-1]).replace("&#215;", "×")
    synname = Name(rawname, auth)
    return synname, accname

def _search(query, progress=None):
    """Do a simple species search on GRIN. Takes the search parameter, e.g.
    'Vicia', and returns a list of Taxon objects.
    """
    query = urlencode({"search":query}).encode('utf-8')
    page = BeautifulSoup(urlopen(TAX_SEARCH_URL, query))
    # Pages for single taxa have h2, a list of results has h1
    if page.h2:
        acctax = _read_sp_page(page)
        accname = acctax.name
        accname.grin_id = acctax.id
        return [(None, accname)]
    
    results = []
    lis = page.ol.findAll("li")
    if progress:
        progress.max(len(lis))
    i = 0
    for li in lis:
        ID = li.a['href'].split("?")[-1]
        if li.a.contents[0].name != "b":
            try:
                namepair = _parse_synonym(li.a)
            except Exception as e:
                warnings.warn("Error parsing name: %s" % e)
                continue
            namepair[1].grin_id = ID
            results.append(namepair)
            continue
        try:
            accname = _parse_name_item(li.a.b)
        except Exception as e:
            warnings.warn("Error parsing name: %s" % e)
            continue
        accname.grin_id = ID
        results.append((None,accname))
        
        i += 1
        if progress and (i % 50 == 0):
            progress(i)

    return results

def get_by_PI(PIcode):
    """Returns basic details of an accession, by searching for its PI number or 
    equivalent code"""
    PIcode = PIcode.lstrip("PI ")
    acc_url = SEARCHb + urllib.urlencode({'accid':"PI "+PIcode})
    return _read_acc_page(acc_url)
    
def get_acc(ID):
    """Returns basic details of an accession, given its numeric ID"""
    return _read_acc_page(ACCb + ID)
    
def _read_acc_page(acc_url):
    acc_page = BeautifulSoup(urlopen(acc_url))
    detail_para = acc_page('p')[0]
    detail_para_s = str(detail_para)
    details = {}
    locn_match = locnre.search(detail_para_s)
    if locn_match:
        details['locn'] = locn_match.group(1)
    details['improvement'] = improvementre.search(detail_para_s).group(1)
    if acc_url.startswith(ACCb):
        details['id'] = acc_url.partition("?")[-1]
    else:
        a = detail_para.findAll("h2")[-1].a
        if a:
            details['id'] = a['href'].partition("?")[-1]			
    return details
    
def get_accs_of_species(sp_id, withUnavailable=True):
    """Given the numeric ID of a species, returns the list of accessions held
    of that species. For each accession, a tuple is returned:
    (numeric id, PI number or equivalent, Accession name)"""
    if withUnavailable:
        listpage = urlopen(TAX_ACCb,"taxno=%s&rownum=0&sort=numb&unavail=off" % sp_id)
    else:
        listpage = urlopen(TAX_ACCb,"taxno=%s&rownum=0&sort=numb" % sp_id)
    listsoup = BeautifulSoup(listpage)
    accs = []
    if not listsoup.ol:
        return []
    for li in listsoup.ol.findAll("li"):
        PInum = li.a.text
        accname = li.text.replace(PInum,"")
        accid = li.a['href'].rpartition("?")[2]
        accs.append((accid, PInum, accname))
    return accs

def _get_distribution(soup):
    try:
        gofrom = soup.find(text="Distributional range").parent.parent
    except AttributeError:    # heading wasn't found
        return {}
    distrib = {}
    while True:
        part = gofrom.findNextSibling("b")
        if not part or part.text not in ("Native:", "Naturalized:", "Cultivated:"):
            break
        countries = set()
        ul = part.findNextSibling("ul")
        for li in ul.findAll("li"):
            # This link is in bold, so it appears in the results if left.
            nse_link = li.find("a", title="Link to NatureServe Explorer")
            if nse_link is not None:
                nse_link.extract()
            countries.update(b.text for b in li.findAll("b")[1:]) # 1st is the continent name
        distrib[part.text.strip(':')] = countries 
        gofrom = ul
    return distrib
    
def _get_species_info(sp_id):
    """Return a dictionary of information for the species with the given
    numeric ID."""
    sp_page = urlopen(TAXON_URL + sp_id)
    soup = BeautifulSoup(sp_page)
    
    info = {'distribution': _get_distribution(soup)}
    
    return info
    
    
def get_obs(ID):
    """Given the numeric ID of an accession, returns a dictionary of the observations made on it."""
    page = BeautifulSoup(urlopen(OBSb + ID))
    obs = defaultdict(list)
    for obtr in page.findAll("tr"):
        if obtr.contents[0].name == "th":
            continue
        obname = obtr.contents[0].a.text
        obs[obname].append((obtr.contents[1].a.text,obtr.contents[1].a['href']))
    return dict(obs)

def _read_sp_page(soup):
    """Read a species page for the name. Returns a GrinTaxon object.
    """
    rank = "sp"
    accname = []
    accauthor = ""
    for chunk in soup.h2.contents[2:]:
        try:
            accname.append(chunk.text)
        except AttributeError:
            chunk = chunk.strip()
            if not chunk:
                continue
            if chunk in ("var.","subsp."):
                accname.append(chunk)
                rank = chunk
            else:
                accauthor = chunk
    accname = " ".join(accname)
    tax = GrinTaxon(accname, accauthor)
    selflink = soup.find("a",title="Jump to images")['href'].partition("#")[0]
    tax.id = selflink.partition("?")[2]
    
    return tax

def fetch(search, progress=None):
    """Fetch taxa matching a search term, e.g. "Restionaceae". Returns a TaxonSet.
    
    The taxa will only have synonymy data, so we only need to make one HTTP
    request.
    """
    by_id = {}
    for synname, accname in _search(search):
        ID = accname.grin_id
        if ID in by_id:
            acctax = by_id[ID]
        else:
            acctax = by_id[ID] = GrinTaxon(accname)
            acctax.id = ID
        
        if synname:
            acctax.othernames.add(synname)
    
    ts = TaxonSet()
    for tax in by_id.values():
        ts.add(tax)
    return ts

class GrinTaxon(Taxon):
    id = ""
    @property
    def url(self):
        return TAXON_URL + self.id
    
    def get_info(self):
        return _get_species_info(self.id)

class GrinTaxaResource(TaxaResource):
    def resolve_name(self, key):
        plainkey = key.plain if hasattr(key, "plain") else key
        names = []
        for synname, accname in _search(plainkey):
            if synname is None:
                names.append((accname, accname))
            elif name.match(key):
                names.append((name, accname))
        return names
    
    # GRIN search accepts wildcards
    wildcard_resolve_name = resolve_name
    
    def get_by_accepted_name(self, name):
        t = GrinTaxon(name)
        t.id = name.grin_id
        return t

_instance = GrinTaxaResource()

resolve_name = _instance.resolve_name
select = _instance.select



from taxonome.external.graphs import MutableGraph, Directed, reversed
from collections import deque

class DiGraph(MutableGraph, Directed):
    """A very simple directional graph class. The implementation is a dict of
    sets, so arbitrary hashable objects can be used for the nodes."""
    class NodesView(MutableGraph.NodesView):
        def __contains__(self, node):
            return node in self.graph.data
            
        def __iter__(self):
            return iter(self.graph.data)
        
        def __len__(self):
            return len(self.graph.data)
        
        def discard(self, node):
            self.graph.data.pop(node, None)
            for _, successors in self.graph.data:
                successors.discard(node)
        
        def add(self, node):
            if node not in self.graph.data:
                self.graph.data[node] = set()
        
    class EdgesView(MutableGraph.EdgesView):
        def __contains__(self, edge):
            tail, head = edge
            return tail in self.graph.data and head in self.graph.data[tail]
            
        def __len__(self):
            return sum(len(links) for _, links in self.graph.data.items())
            
        def __iter__(self):
            for tail, links in self.graph.data.items():
                for head in links:
                    yield (tail, head)
                    
        def __getitem__(self, edge):
            if edge in self:
                return 1
            raise KeyError(edge)
        
        def __setitem__(self, edge, value):
            if value != 1:
                raise ValueError("Edge labels are not supported.")
            tail, head = edge
            self.graph.nodes.add(tail)
            self.graph.nodes.add(head)
            self.graph.data[tail].add(head)
            
        def __delitem__(self, edge):
            if edge not in self:
                raise KeyError(edge)
            tail, head = edge
            self.graph.data[tail].discard(head)
            
    def __init__(self):
        super(DiGraph, self).__init__()
        self.data = {}
    
    def __contains__(self, key):
        return key in self.data
        
    @classmethod
    def build(cls, edges, nodes=[]):
        g = cls()
        g.data = {n: set() for n in nodes}
        for t,h in edges:
            g.edges.add(t,h)
        return g
    
    def successors(self, node, depth_first=False):
        """Iterate over all successors of the given node.
        
        By default, it will look breadth first. If depth_first is True, it will
        look depth first.
        """
        stq = deque()
        push = stq.extend if depth_first else stq.extendleft
        push(self.data[node])
        already_done = set([node])
        while stq:
            node = stq.pop()
            if node not in already_done:
                yield node
                push(self.data[node])
                already_done.add(node)

    def is_successor(self, predecessor, subject, depth_first=False):
        """Return True if subject is a successor of predecessor.
        
        By default, it will search breadth first. This can be changed by setting
        depth_first to True."""
        if subject in self.data[predecessor]:
            return True
        
        stq = deque()
        push = stq.extend if depth_first else stq.extendleft
        push(self.data[predecessor])
        already_done = set([predecessor])
        while stq:
            predecessor = stq.pop()
            if predecessor not in already_done:
                links = self.data[predecessor]
                if subject in links:
                    return True
                push(links)
                already_done.add(predecessor)
        
        return False
    
    def immediate_predecessors(self, node):
        """Iterate over the immediate predecessors of a given node."""
        for x, xchildren in self.data.items():
            if node in xchildren:
                yield x
    
    def predecessors(self, node):
        """Iterate over all predecessors of a node. Note that this is slower
        than iterating over successors."""
        stq = deque(self.immediate_predecessors(node))
        already_done = set([node])
        while stq:
            nextnode = stq.pop()
            if nextnode not in already_done:
                yield nextnode
                stq.extend(self.immediate_predecessors(nextnode))
                already_done.add(nextnode)

# We need to alias these classes here for pickle to accept DiGraph instances.
NodesView = DiGraph.NodesView
EdgesView = DiGraph.EdgesView



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
import taxonome
from taxonome.taxa import file_csv
import unittest
from os import remove

Name = taxonome.Name    

class ElephantTest(unittest.TestCase):
    """Tests both taxa and distribution."""
    def setUp(self):
        self.ts = taxonome.TaxonSet()
        a_elephant = taxonome.Taxon("Elephas maximus","L.")
        name_ix = taxonome.regions.names
        a_elephant.distribution = {name_ix[n] for n in \
                                    ["India", "Nepal","Thailand","Bhutan"]}
        self.ts.add(a_elephant)
        
        forest_elephant = taxonome.Taxon(Name("Loxodonta cyclotis","Matchie"))
        forest_elephant.distribution = {name_ix["Congo"]}
        self.ts.add(forest_elephant)
        
    def testTS(self):
        assert len(self.ts) == 2
        assert Name("Elephas maximus","L") in self.ts
        assert self.ts["Elephas maximus"].name == Name("Elephas maximus","L")
        assert "Loxodonta cyclotis" in self.ts
    
    def testDistrib(self):
        a_elephant = self.ts["Elephas maximus"]
        world = taxonome.regions.world
        trop_asia = taxonome.regions.tdwg["4"]     # 4 is Tropical Asia, TDWG L1
        africa = taxonome.regions.tdwg["2"]        # 2 is Africa, TDWG L1
        assert africa == taxonome.regions.names["Africa"]
        assert world.includes(trop_asia, a_elephant.distribution)
        assert not world.includes(africa, a_elephant.distribution)
        
        forest_elephant = self.ts[Name("Loxodonta cyclotis", "Matchie")]
        assert world.includes(africa, forest_elephant.distribution)
        denmark = taxonome.regions.names["Denmark"]
        assert not world.includes(denmark, forest_elephant.distribution)

if __name__ == "__main__":
    unittest.main()



pair_fail_msg = ("Testing function {0}\n\n"
                "In:\n"
                " {1!r}\n"
                "Expected:\n"
                " {2!r}\n"
                "Got:\n"
                " {3!r}\n")
def check_func(func, pairs):
    """Utility function for the common case of checking a function with a
    sequence of input/output pairs.
    
    Parameters
    ----------
    func : callable
      The function to be tested. Should accept a single argument.
    pairs : iterable
      A list of (*input, expected_output) tuples.
    
    Returns
    -------
    None. Raises an AssertionError if any output does not match the expected
    value.
    """
    for *inp, expected in pairs:
        out = func(*inp)
        assert out == expected, \
                        pair_fail_msg.format(func.__name__, inp, expected, out)



from unittest import TestCase
from taxonome.utils import qgram, testing

# Shorthand
S, E = qgram.START, qgram.END
tests = [("abcdef", {S*2+"a", S+"ab", "abc", "bcd", "cde", "def", "ef"+E,
                        "f"+E*2})]

tests4 = [("abcdef", {S*3+"a", S*2+"ab", S+"abc", "abcd", "bcde", "cdef", 
                        "def"+E, "ef"+E*2, "f"+E*3})
         ]
q4grams = lambda s: qgram.qgrams(s, q=4)

def test_qgram():
    testing.check_func(q4grams, tests4)
    testing.check_func(qgram.qgrams, tests)

# Check we have Python 3 style divsion:
assert 1/2 == 0.5

match_tests = [("Trisetum preslei", "Trisetum preslii", 5/6)]

def test_compare_strings():
    testing.check_func(qgram.compare_strings, match_tests)

class TestQGramIndex(TestCase):
    def test_lookup(self):
        qindex = qgram.QGramIndex()
        qindex.add("Hilaria swalleni")
        res, score = qindex.best_match("Hilaria swallenii") # N.B. Double i
        self.assertEqual(res, "Hilaria swalleni")
        self.assertGreater(score, 0.85)



# -*- coding: utf-8 -*-
"""This contains data for the TDWG world regions. Do not use this module
directly, but instead do:
from taxonome import regions
regions.names["China"]
regions.tdwg["NET"]
regions.tdwg_levels[1]["1"]

If load-tdwg-regions is false in settings, these can be loaded with:
regions.load_tdwg()

Data is as downloaded from:
http://www.tdwg.org/standards/109
(Edition 2)

It also uses an ISO index, available from:
http://www.iso.org/iso/country_codes/iso_3166_code_lists.htm
"""
data = [\
# TDWG level 1:
('1','Europe',None,'!!WORLD', None),
('2','Africa',None,'!!WORLD', None),
('3','Asia-Temperate',None,'!!WORLD', None),
('4','Asia-Tropical',None,'!!WORLD', None),
('5','Australasia',None,'!!WORLD', None),
('6','Pacific',None,'!!WORLD', None),
('7','Northern America',None,'!!WORLD', None),
('8','Southern America',None,'!!WORLD', None),
('9','Antarctic',None,'!!WORLD', None),

# TDWG level 2:
('10','Northern Europe',None,'1', None),
('11','Middle Europe',None,'1', None),
('12','Southwestern Europe',None,'1', None),
('13','Southeastern Europe',None,'1', None),
('14','Eastern Europe',None,'1', None),
('20','Northern Africa',None,'2', None),
('21','Macaronesia',None,'2', None),
('22','West Tropical Africa',None,'2', None),
('23','West-Central Tropical Africa',None,'2', None),
('24','Northeast Tropical Africa',None,'2', None),
('25','East Tropical Africa',None,'2', None),
('26','South Tropical Africa',None,'2', None),
('27','Southern Africa',None,'2', None),
('28','Middle Atlantic Ocean',None,'2', None),
('29','Western Indian Ocean',None,'2', None),
('30','Siberia',None,'3', None),
('31','Russian Far East',None,'3', None),
('32','Middle Asia',None,'3', None),
('33','Caucasus',None,'3', None),
('34','Western Asia',None,'3', None),
('35','Arabian Peninsula',None,'3', None),
('36','China','CN','3', None),
('37','Mongolia','MN','3', None),
('38','Eastern Asia',None,'3', None),
('40','Indian Subcontinent',None,'4', None),
('41','Indo-China',None,'4', None),
('42','Malesia',None,'4', None),
('43','Papuasia',None,'4', None),
('50','Australia','AU','5', None),
('51','New Zealand','NZ','5', None),
('60','Southwestern Pacific',None,'6', None),
('61','South-Central Pacific',None,'6', None),
('62','Northwestern Pacific',None,'6', None),
('63','North-Central Pacific',None,'6', None),
('70','Subarctic America',None,'7', None),
('71','Western Canada',None,'7', None),
('72','Eastern Canada',None,'7', None),
('73','Northwestern U.S.A.',None,'7', None),
('74','North-Central U.S.A.',None,'7', None),
('75','Northeastern U.S.A.',None,'7', None),
('76','Southwestern U.S.A.',None,'7', None),
('77','South-Central U.S.A.',None,'7', None),
('78','Southeastern U.S.A.',None,'7', None),
('79','Mexico','MX','7', None),
('80','Central America',None,'8', None),
('81','Caribbean',None,'8', None),
('82','Northern South America',None,'8', None),
('83','Western South America',None,'8', None),
('84','Brazil','BR','8', None),
('85','Southern South America',None,'8', None),
('90','Subantarctic Islands',None,'9', None),
('91','Antarctic Continent',None,'9', None),

# TDWG level 3:
('ABT','Alberta',None,'71', None),
('AFG','Afghanistan','AF','34', None),
('AGE','Argentina Northeast',None,'85', None),
('AGS','Argentina South',None,'85', None),
('AGW','Argentina Northwest',None,'85', None),
('ALA','Alabama',None,'78', None),
('ALB','Albania','AL','13', None),
('ALD','Aldabra',None,'29', None),
('ALG','Algeria','DZ','20', None),
('ALT','Altay',None,'30', None),
('ALU','Aleutian Is.',None,'70', None),
('AMU','Amur',None,'31', None),
('AND','Andaman Is.',None,'41', None),
('ANG','Angola','AO','26', None),
('ANT','Antarctica','AQ','91', None),
('ARI','Arizona',None,'76', None),
('ARK','Arkansas',None,'78', None),
('ARU','Aruba','AW','81', None),
('ASC','Ascension',None,'28', None),
('ASK','Alaska',None,'70', None),
('ASP','Amsterdam-St.Paul Is.',None,'90', None),
('ASS','Assam',None,'40', None),
('ATP','Antipodean Is.',None,'51', None),
('AUT','Austria','AT','11', None),
('AZO','Azores',None,'21', None),
('BAH','Bahamas','BS','81', None),
('BAL','Baleares',None,'12', None),
('BAN','Bangladesh','BD','40', None),
('BEN','Benin','BJ','22', None),
('BER','Bermuda','BM','81', None),
('BGM','Belgium','BE','11', None),
('BIS','Bismarck Archipelago',None,'43', None),
('BKN','Burkina',None,'22', None),
('BLR','Belarus','BY','14', None),
('BLT','Baltic States',None,'14', None),
('BLZ','Belize','BZ','80', None),
('BOL','Bolivia',None,'83', None),
('BOR','Borneo',None,'42', None),
('BOT','Botswana','BW','27', None),
('BOU','Bouvet I.',None,'90', None),
('BRC','British Columbia',None,'71', None),
('BRY','Buryatiya',None,'30', None),
('BUL','Bulgaria','BG','13', None),
('BUR','Burundi','BI','23', None),
('BZC','Brazil West-Central',None,'84', None),
('BZE','Brazil Northeast',None,'84', None),
('BZL','Brazil Southeast',None,'84', None),
('BZN','Brazil North',None,'84', None),
('BZS','Brazil South',None,'84', None),
('CAB','Cabinda',None,'23', None),
('CAF','Central African Republic','CF','23', None),
('CAL','California',None,'76', None),
('CAY','Cayman Is.',None,'81', None),
('CBD','Cambodia','KH','41', None),
('CGS','Chagos Archipelago',None,'29', None),
('CHA','Chad','TD','24', None),
('CHC','China South-Central',None,'36', None),
('CHH','Hainan',None,'36', None),
('CHI','Inner Mongolia',None,'36', None),
('CHM','Manchuria',None,'36', None),
('CHN','China North-Central',None,'36', None),
('CHQ','Qinghai',None,'36', None),
('CHS','China Southeast',None,'36', None),
('CHT','Tibet',None,'36', None),
('CHX','Xinjiang',None,'36', None),
('CKI','Cocos (Keeling) Is.',None,'42', None),
('CLC','Chile Central',None,'85', None),
('CLM','Colombia','CO','83', None),
('CLN','Chile North',None,'85', None),
('CLS','Chile South',None,'85', None),
('CMN','Cameroon','CM','23', None),
('CNT','Connecticut',None,'75', None),
('CNY','Canary Is.',None,'21', None),
('COL','Colorado',None,'73', None),
('COM','Comoros','KM','29', None),
('CON','Congo','CG','23', None),
('COO','Cook Is.',None,'61', None),
('COR','Corse',None,'12', None),
('COS','Costa Rica','CR','80', None),
('CPI','Central American Pacific Is.',None,'80', None),
('CPP','Cape Provinces',None,'27', None),
('CPV','Caprivi Strip',None,'27', None),
('CRL','Caroline Is.',None,'62', None),
('CRZ','Crozet Is.',None,'90', None),
('CTA','Chita',None,'30', None),
('CTM','Chatham Is.',None,'51', None),
('CUB','Cuba','CU','81', None),
('CVI','Cape Verde','CV','21', None),
('CYP','Cyprus','CY','34', None),
('CZE','Czechoslovakia',None,'11', None),
('DEL','Delaware',None,'78', None),
('DEN','Denmark','DK','10', None),
('DJI','Djibouti','DJ','24', None),
('DOM','Dominican Republic','DO','81', None),
('DSV','Desventurados Is.',None,'85', None),
('EAI','East Aegean Is.',None,'34', None),
('EAS','Easter Is.',None,'61', None),
('ECU','Ecuador','EC','83', None),
('EGY','Egypt','EG','20', None),
('EHM','East Himalaya',None,'40', None),
('ELS','El Salvador','SV','80', None),
('EQG','Equatorial Guinea','GQ','23', None),
('ERI','Eritrea','ER','24', None),
('ETH','Ethiopia','ET','24', None),
('FAL','Falkland Is.',None,'90', None),
('FIJ','Fiji','FJ','60', None),
('FIN','Finland','FI','10', None),
('FLA','Florida',None,'78', None),
('FOR','Føroyar',None,'10', None),
('FRA','France','FR','12', None),
('FRG','French Guiana','GF','82', None),
('GAB','Gabon','GA','23', None),
('GAL','Galápagos',None,'83', None),
('GAM','Gambia, The',None,'22', None),
('GEO','Georgia','GE','78', None),
('GER','Germany','DE','11', None),
('GGI','Gulf of Guinea Is.',None,'23', None),
('GHA','Ghana','GH','22', None),
('GIL','Gilbert Is.',None,'60', None),
('GNB','Guinea-Bissau','GW','22', None),
('GNL','Greenland','GL','70', None),
('GRB','Great Britain',None,'10', None),
('GRC','Greece','GR','13', None),
('GST','Gulf States',None,'35', None),
('GUA','Guatemala','GT','80', None),
('GUI','Guinea','GN','22', None),
('GUY','Guyana','GY','82', None),
('HAI','Haiti','HT','81', None),
('HAW','Hawaii',None,'63', None),
('HBI','Howland-Baker Is.',None,'60', None),
('HMD','Heard-McDonald Is.',None,'90', None),
('HON','Honduras','HN','80', None),
('HUN','Hungary','HU','11', None),
('ICE','Iceland','IS','10', None),
('IDA','Idaho',None,'73', None),
('ILL','Illinois',None,'74', None),
('IND','India','IN','40', None),
('INI','Indiana',None,'75', None),
('IOW','Iowa',None,'74', None),
('IRE','Ireland','IE','10', None),
('IRK','Irkutsk',None,'30', None),
('IRN','Iran',None,'34', None),
('IRQ','Iraq','IQ','34', None),
('ITA','Italy','IT','13', None),
('IVO','Ivory Coast',None,'22', None),
('JAM','Jamaica','JM','81', None),
('JAP','Japan','JP','38', None),
('JAW','Jawa',None,'42', None),
('JNF','Juan Fernández Is.',None,'85', None),
('KAM','Kamchatka',None,'31', None),
('KAN','Kansas',None,'74', None),
('KAZ','Kazakhstan','KZ','32', None),
('KEG','Kerguelen',None,'90', None),
('KEN','Kenya','KE','25', None),
('KER','Kermadec Is.',None,'51', None),
('KGZ','Kirgizistan',None,'32', None),
('KHA','Khabarovsk',None,'31', None),
('KOR','Korea',None,'38', None),
('KRA','Krasnoyarsk',None,'30', None),
('KRI','Kriti',None,'13', None),
('KRY','Krym',None,'14', None),
('KTY','Kentucky',None,'78', None),
('KUR','Kuril Is.',None,'31', None),
('KUW','Kuwait','KW','35', None),
('KZN','Kazan-retto',None,'38', None),
('LAB','Labrador',None,'72', None),
('LAO','Laos',None,'41', None),
('LBR','Liberia','LR','22', None),
('LBS','Lebanon-Syria',None,'34', None),
('LBY','Libya',None,'20', None),
('LDV','Laccadive Is.',None,'40', None),
('LEE','Leeward Is.',None,'81', None),
('LES','Lesotho','LS','27', None),
('LIN','Line Is.',None,'61', None),
('LOU','Louisiana',None,'78', None),
('LSI','Lesser Sunda Is.',None,'42', None),
('MAG','Magadan',None,'31', None),
('MAI','Maine',None,'75', None),
('MAN','Manitoba',None,'71', None),
('MAQ','Macquarie Is.',None,'90', None),
('MAS','Masachusettes',None,'75', None),
('MAU','Mauritius','MU','29', None),
('MCI','Mozambique Channel Is.',None,'29', None),
('MCS','Marcus I.',None,'62', None),
('MDG','Madagascar','MG','29', None),
('MDR','Madeira',None,'21', None),
('MDV','Maldives','MV','40', None),
('MIC','Michigan',None,'75', None),
('MIN','Minnesota',None,'74', None),
('MLI','Mali','ML','22', None),
('MLW','Malawi','MW','26', None),
('MLY','Malaya',None,'42', None),
('MNT','Montana',None,'73', None),
('MOL','Maluku',None,'42', None),
('MON','Mongolia','MN','37', None),
('MOR','Morocco','MA','20', None),
('MOZ','Mozambique','MZ','26', None),
('MPE','Marion-Prince Edward Is.',None,'90', None),
('MRN','Marianas',None,'62', None),
('MRQ','Marquesas',None,'61', None),
('MRS','Marshall Is.',None,'62', None),
('MRY','Maryland',None,'78', None),
('MSI','Mississippi',None,'78', None),
('MSO','Missouri',None,'74', None),
('MTN','Mauritania','MR','22', None),
('MXC','Mexico Central',None,'79', None),
('MXE','Mexico Northeast',None,'79', None),
('MXG','Mexico Gulf',None,'79', None),
('MXI','Mexican Pacific Is.',None,'79', None),
('MXN','Mexico Northwest',None,'79', None),
('MXS','Mexico Southwest',None,'79', None),
('MXT','Mexico Southeast',None,'79', None),
('MYA','Myanmar','MM','41', None),
('NAM','Namibia','NA','27', None),
('NAT','KwaZulu-Natal',None,'27', None),
('NBR','New Brunswick',None,'72', None),
('NCA','North Carolina',None,'78', None),
('NCB','Nicobar Is.',None,'41', None),
('NCS','North Caucasus',None,'33', None),
('NDA','North Dakota',None,'74', None),
('NEB','Nebraska',None,'74', None),
('NEP','Nepal','NP','40', None),
('NET','Netherlands','NL','11', None),
('NEV','Nevada',None,'76', None),
('NFK','Norfolk Is.',None,'50', None),
('NFL','Newfoundland',None,'72', None),
('NGA','Nigeria','NG','22', None),
('NGR','Niger','NE','22', None),
('NIC','Nicaragua','NI','80', None),
('NLA','Netherlands Antilles','AN','81', None),
('NNS','Nansei-shoto',None,'38', None),
('NOR','Norway','NO','10', None),
('NRU','Nauru','NR','60', None),
('NSC','Nova Scotia',None,'72', None),
('NSW','New South Wales',None,'50', None),
('NTA','Northern Territory',None,'50', None),
('NUE','Niue','NU','60', None),
('NUN','Nunavut',None,'70', None),
('NWC','New Caledonia','NC','60', None),
('NWG','New Guinea',None,'43', None),
('NWH','New Hampshire',None,'75', None),
('NWJ','New Jersey',None,'75', None),
('NWM','New Mexico',None,'77', None),
('NWT','Northwest Territories',None,'70', None),
('NWY','New York',None,'75', None),
('NZN','New Zealand North',None,'51', None),
('NZS','New Zealand South',None,'51', None),
('OFS','Free State',None,'27', None),
('OGA','Ogasawara-shoto',None,'38', None),
('OHI','Ohio',None,'75', None),
('OKL','Oklahoma',None,'74', None),
('OMA','Oman','OM','35', None),
('ONT','Ontario',None,'72', None),
('ORE','Oregon',None,'73', None),
('PAK','Pakistan','PK','40', None),
('PAL','Palestine',None,'34', None),
('PAN','Panamá',None,'80', None),
('PAR','Paraguay','PY','85', None),
('PEI','Prince Edward I.',None,'72', None),
('PEN','Pennsylvania',None,'75', None),
('PER','Peru','PE','83', None),
('PHI','Philippines','PH','42', None),
('PHX','Phoenix Is.',None,'60', None),
('PIT','Pitcairn Is.',None,'61', None),
('POL','Poland','PL','11', None),
('POR','Portugal','PT','12', None),
('PRM','Primorye',None,'31', None),
('PUE','Puerto Rico','PR','81', None),
('QLD','Queensland',None,'50', None),
('QUE','Québec',None,'72', None),
('REU','Réunion','RE','29', None),
('RHO','Rhode I.',None,'75', None),
('ROD','Rodrigues',None,'29', None),
('ROM','Romania','RO','13', None),
('RUC','Central European Russia',None,'14', None),
('RUE','East European Russia',None,'14', None),
('RUN','North European Russia',None,'14', None),
('RUS','South European Russia',None,'14', None),
('RUW','Northwest European Russia',None,'14', None),
('RWA','Rwanda','RW','23', None),
('SAK','Sakhalin',None,'31', None),
('SAM','Samoa','WS','60', None),
('SAR','Sardegna',None,'12', None),
('SAS','Saskatchewan',None,'71', None),
('SAU','Saudi Arabia','SA','35', None),
('SCA','South Carolina',None,'78', None),
('SCI','Society Is.',None,'61', None),
('SCS','South China Sea',None,'41', None),
('SCZ','Santa Cruz Is.',None,'60', None),
('SDA','South Dakota',None,'74', None),
('SEL','Selvagens',None,'21', None),
('SEN','Senegal','SN','22', None),
('SEY','Seychelles','SC','29', None),
('SGE','South Georgia',None,'90', None),
('SIC','Sicilia',None,'13', None),
('SIE','Sierra Leone','SL','22', None),
('SIN','Sinai',None,'34', None),
('SOA','South Australia',None,'50', None),
('SOC','Socotra',None,'24', None),
('SOL','Solomon Is.',None,'43', None),
('SOM','Somalia','SO','24', None),
('SPA','Spain','ES','12', None),
('SRL','Sri Lanka','LK','40', None),
('SSA','South Sandwich Is.',None,'90', None),
('STH','St.Helena',None,'28', None),
('SUD','Sudan','SD','24', None),
('SUL','Sulawesi',None,'42', None),
('SUM','Sumatera',None,'42', None),
('SUR','Suriname','SR','82', None),
('SVA','Svalbard',None,'10', None),
('SWC','Southwest Caribbean',None,'81', None),
('SWE','Sweden','SE','10', None),
('SWI','Switzerland','CH','11', None),
('SWZ','Swaziland','SZ','27', None),
('TAI','Taiwan',None,'38', None),
('TAN','Tanzania',None,'25', None),
('TAS','Tasmania',None,'50', None),
('TCI','Turks-Caicos Is.',None,'81', None),
('TCS','Transcaucasus',None,'33', None),
('TDC','Tristan da Cunha',None,'90', None),
('TEN','Tennessee',None,'78', None),
('TEX','Texas',None,'77', None),
('THA','Thailand','TH','41', None),
('TKM','Turkmenistan','TM','32', None),
('TOG','Togo','TG','22', None),
('TOK','Tokelau-Manihiki',None,'60', None),
('TON','Tonga','TO','60', None),
('TRT','Trinidad-Tobago',None,'81', None),
('TUA','Tuamotu',None,'61', None),
('TUB','Tubuai Is.',None,'61', None),
('TUE','Turkey-in-Europe',None,'13', None),
('TUN','Tunisia','TN','20', None),
('TUR','Turkey','TR','34', None),
('TUV','Tuvalu','TV','60', None),
('TVA','Tuva',None,'30', None),
('TVL','Northern Provinces',None,'27', None),
('TZK','Tadzhikistan',None,'32', None),
('UGA','Uganda','UG','25', None),
('UKR','Ukraine','UA','14', None),
('URU','Uruguay','UY','85', None),
('UTA','Utah',None,'76', None),
('UZB','Uzbekistan','UZ','32', None),
('VAN','Vanuatu','VU','60', None),
('VEN','Venezuela',None,'82', None),
('VER','Vermont',None,'75', None),
('VIC','Victoria',None,'50', None),
('VIE','Vietnam',None,'41', None),
('VNA','Venezuelan Antilles',None,'81', None),
('VRG','Virginia',None,'78', None),
('WAK','Wake I.',None,'62', None),
('WAL','Wallis-Futuna Is.',None,'60', None),
('WAS','Washington',None,'73', None),
('WAU','Western Australia',None,'50', None),
('WDC','District of Columbia',None,'78', None),
('WHM','West Himalaya',None,'40', None),
('WIN','Windward Is.',None,'81', None),
('WIS','Wisconsin',None,'74', None),
('WSA','Western Sahara','EH','20', None),
('WSB','West Siberia',None,'30', None),
('WVA','West Virginia',None,'75', None),
('WYO','Wyoming',None,'73', None),
('XMS','Christmas I.',None,'42', None),
('YAK','Yakutskiya',None,'30', None),
('YEM','Yemen','YE','35', None),
('YUG','Yugoslavia',None,'13', None),
('YUK','Yukon',None,'70', None),
('ZAI','Zaire',None,'23', None),
('ZAM','Zambia','ZM','26', None),
('ZIM','Zimbabwe','ZW','26', None),

# TDWG level 4:
('ABT-OO',"Alberta",None,'ABT',''),
('AFG-OO',"Afghanistan",'AF','AFG',''),
('AGE-BA',"Buenos Aires",None,'AGE',''),
('AGE-CH',"Chaco",None,'AGE',''),
('AGE-CN',"Corrientes",None,'AGE',''),
('AGE-CO',"Córdoba",None,'AGE',''),
('AGE-DF',"Argentina Distrito Federal",None,'AGE',''),
('AGE-ER',"Entre Ríos",None,'AGE',''),
('AGE-FO',"Formosa",None,'AGE',''),
('AGE-LP',"La Pampa",None,'AGE',''),
('AGE-MI',"Misiones",None,'AGE',''),
('AGS-CB',"Chubut",None,'AGS',''),
('AGS-NE',"Neuquén",None,'AGS',''),
('AGS-RN',"Rio Negro",None,'AGS',''),
('AGS-SC',"Santa Cruz",None,'AGS',''),
('AGS-SF',"Santa Fé",None,'AGS',''),
('AGS-TF',"Tierra del Fuego (Argentina)",None,'AGS',''),
('AGW-CA',"Catamarca",None,'AGW',''),
('AGW-JU',"Jujuy",None,'AGW',''),
('AGW-LR',"La Rioja",None,'AGW',''),
('AGW-ME',"Mendoza",None,'AGW',''),
('AGW-SA',"Salta",None,'AGW',''),
('AGW-SE',"Santiago del Estero",None,'AGW',''),
('AGW-SJ',"San Juan",None,'AGW',''),
('AGW-SL',"San Luis",None,'AGW',''),
('AGW-TU',"Tucuman",None,'AGW',''),
('ALA-OO',"Alabama",None,'ALA',''),
('ALB-OO',"Albania",'AL','ALB',''),
('ALD-OO',"Aldabra",None,'ALD',''),
('ALG-OO',"Algeria",'DZ','ALG',''),
('ALT-OO',"Altay",None,'ALT',''),
('ALU-OO',"Aleutian Is.",None,'ALU',''),
('AMU-OO',"Amur",None,'AMU',''),
('AND-AN',"Andaman Is.",None,'AND',''),
('AND-CO',"Coco Is.",None,'AND',''),
('ANG-OO',"Angola",'AO','ANG',''),
('ANT-OO',"Antarctica",'AQ','ANT',''),
('ARI-OO',"Arizona",None,'ARI',''),
('ARK-OO',"Arkansas",None,'ARK',''),
('ARU-OO',"Aruba",'AW','ARU',''),
('ASC-OO',"Ascension",None,'ASC',''),
('ASK-OO',"Alaska",None,'ASK',''),
('ASP-OO',"Amsterdam-St.Paul Is.",None,'ASP',''),
('ASS-AS',"Assam",None,'ASS',''),
('ASS-MA',"Manipur",None,'ASS',''),
('ASS-ME',"Meghalaya",None,'ASS',''),
('ASS-MI',"Mizoram",None,'ASS',''),
('ASS-NA',"Nagaland",None,'ASS',''),
('ASS-TR',"Tripura",None,'ASS',''),
('ATP-OO',"Antipodean Is.",None,'ATP',''),
('AUT-AU',"Austria",'AT','AUT',''),
('AUT-LI',"Liechtenstein",'LI','AUT',''),
('AZO-OO',"Açôres",None,'AZO',''),
('BAH-OO',"Bahamas",'BS','BAH',''),
('BAL-OO',"Baleares",None,'BAL',''),
('BAN-OO',"Bangladesh",'BD','BAN',''),
('BEN-OO',"Benin",'BJ','BEN',''),
('BER-OO',"Bermuda",'BM','BER',''),
('BGM-BE',"Belgium",'BE','BGM',''),
('BGM-LU',"Luxembourg",'LU','BGM',''),
('BIS-OO',"Bismarck Archipelago",None,'BIS',''),
('BKN-OO',"Burkina",None,'BKN',''),
('BLR-OO',"Belarus",'BY','BLR',''),
('BLT-ES',"Estonia",'EE','BLT',''),
('BLT-KA',"Kaliningrad",None,'BLT',''),
('BLT-LA',"Latvia",'LV','BLT',''),
('BLT-LI',"Lithuania",'LT','BLT',''),
('BLZ-OO',"Belize",'BZ','BLZ',''),
('BOL-OO',"Bolivia",None,'BOL',''),
('BOR-BR',"Brunei",None,'BOR',''),
('BOR-KA',"Kalimantan",None,'BOR',''),
('BOR-SB',"Sabah",None,'BOR',''),
('BOR-SR',"Sarawak",None,'BOR',''),
('BOT-OO',"Botswana",'BW','BOT',''),
('BOU-OO',"Bouvet I.",None,'BOU',''),
('BRC-OO',"British Columbia",None,'BRC',''),
('BRY-OO',"Buryatiya",None,'BRY',''),
('BUL-OO',"Bulgaria",'BG','BUL',''),
('BUR-OO',"Burundi",'BI','BUR',''),
('BZC-DF',"Brazilia Distrito Federal",None,'BZC',''),
('BZC-GO',"Goiás",None,'BZC',''),
('BZC-MS',"Mato Grosso do Sul",None,'BZC',''),
('BZC-MT',"Mato Grosso",None,'BZC',''),
('BZE-AL',"Alagoas",None,'BZE',''),
('BZE-BA',"Bahia",None,'BZE',''),
('BZE-CE',"Ceará",None,'BZE',''),
('BZE-FN',"Fernando de Noronha",None,'BZE',''),
('BZE-MA',"Maranhao",None,'BZE',''),
('BZE-PB',"Paraíba",None,'BZE',''),
('BZE-PE',"Pernambuco",None,'BZE',''),
('BZE-PI',"Piauí",None,'BZE',''),
('BZE-RN',"Rio Grande do Norte",None,'BZE',''),
('BZE-SE',"Sergipe",None,'BZE',''),
('BZL-ES',"Espirito Santo",None,'BZL',''),
('BZL-MG',"Minas Gerais",None,'BZL',''),
('BZL-RJ',"Rio de Janeiro",None,'BZL',''),
('BZL-SP',"São Paulo",None,'BZL',''),
('BZL-TR',"Trindade",None,'BZL',''),
('BZN-AC',"Acre",None,'BZN',''),
('BZN-AM',"Amazonas",None,'BZN',''),
('BZN-AP',"Amapá",None,'BZN',''),
('BZN-PA',"Pará",None,'BZN',''),
('BZN-RM',"Roraima",None,'BZN',''),
('BZN-RO',"Rondônia",None,'BZN',''),
('BZN-TO',"Tocantins",None,'BZN',''),
('BZS-PR',"Paraná",None,'BZS',''),
('BZS-RS',"Rio Grande do Sul",None,'BZS',''),
('BZS-SC',"Santa Catarina",None,'BZS',''),
('CAB-OO',"Cabinda",None,'CAB',''),
('CAF-OO',"Central African Republic",'CF','CAF',''),
('CAL-OO',"California",None,'CAL',''),
('CAY-OO',"Cayman Is.",None,'CAY',''),
('CBD-OO',"Cambodia",'KH','CBD',''),
('CGS-OO',"Chagos Archipelago",None,'CGS',''),
('CHA-OO',"Chad",'TD','CHA',''),
('CHC-GZ',"Guizhou",None,'CHC',''),
('CHC-HU',"Hubei",None,'CHC',''),
('CHC-SC',"Sichuan",None,'CHC',''),
('CHC-YN',"Yunnan",None,'CHC',''),
('CHH-OO',"Hainan",None,'CHH',''),
('CHI-NM',"Nei Mongol",None,'CHI',''),
('CHI-NX',"Ningxia",None,'CHI',''),
('CHM-HJ',"Heilongjiang",None,'CHM',''),
('CHM-JL',"Jilin",None,'CHM',''),
('CHM-LN',"Liaoning",None,'CHM',''),
('CHN-BJ',"Beijing",None,'CHN',''),
('CHN-GS',"Gansu",None,'CHN',''),
('CHN-HB',"Hebei",None,'CHN',''),
('CHN-SA',"Shaanxi",None,'CHN',''),
('CHN-SD',"Shandong",None,'CHN',''),
('CHN-SX',"Shanxi",None,'CHN',''),
('CHN-TJ',"Tianjin",None,'CHN',''),
('CHQ-OO',"Qinghai",None,'CHQ',''),
('CHS-AH',"Anhui",None,'CHS',''),
('CHS-FJ',"Fujian",None,'CHS',''),
('CHS-GD',"Guangdong",None,'CHS',''),
('CHS-GX',"Guangxi",None,'CHS',''),
('CHS-HE',"Henan",None,'CHS',''),
('CHS-HK',"Hong Kong",'HK','CHS',''),
('CHS-HN',"Hunan",None,'CHS',''),
('CHS-JS',"Jiangsu",None,'CHS',''),
('CHS-JX',"Jiangxi",None,'CHS',''),
('CHS-KI',"Kin-Men",None,'CHS',''),
('CHS-MA',"Macau",None,'CHS',''),
('CHS-MP',"Ma-tsu-Pai-chúan",None,'CHS',''),
('CHS-SH',"Shanghai",None,'CHS',''),
('CHS-ZJ',"Zhejiang",None,'CHS',''),
('CHT-OO',"Tibet",None,'CHT','Sube del nivel 4 al nivel 3. El código en el libro es CHT-XI.'),
('CHX-OO',"Xinjiang",None,'CHX',''),
('CKI-OO',"Cocos (Keeling) Is.",None,'CKI',''),
('CLC-BI',"Biobío",None,'CLC',''),
('CLC-CO',"Coquimbo",None,'CLC',''),
('CLC-LA',"La Araucania",None,'CLC',''),
('CLC-MA',"Maule",None,'CLC',''),
('CLC-OH',"O'Higgins",None,'CLC',''),
('CLC-SA',"Santiago",None,'CLC',''),
('CLC-VA',"Valparaíso",None,'CLC',''),
('CLM-OO',"Colombia",'CO','CLM',''),
('CLN-AN',"Antofagasta",None,'CLN',''),
('CLN-AT',"Atacama",None,'CLN',''),
('CLN-TA',"Tarapaca",None,'CLN',''),
('CLS-AI',"Aisén",None,'CLS',''),
('CLS-LL',"Los Lagos",None,'CLS',''),
('CLS-MG',"Magellanes",None,'CLS',''),
('CMN-OO',"Cameroon",'CM','CMN',''),
('CNT-OO',"Connecticut",None,'CNT',''),
('CNY-OO',"Canary Is.",None,'CNY',''),
('COL-OO',"Colorado",None,'COL',''),
('COM-CO',"Comoros",'KM','COM',''),
('COM-MA',"Mayotte",'YT','COM',''),
('CON-OO',"Congo",'CG','CON',''),
('COO-OO',"Cook Is.",None,'COO',''),
('COR-OO',"Corse",None,'COR',''),
('COS-OO',"Costa Rica",'CR','COS',''),
('CPI-CL',"Clipperton I.",None,'CPI',''),
('CPI-CO',"Cocos I.",None,'CPI',''),
('CPI-MA',"Malpelo I.",None,'CPI',''),
('CPP-EC',"Eastern Cape Province",None,'CPP',''),
('CPP-NC',"Northern Cape Province",None,'CPP',''),
('CPP-WC',"Western Cape Province",None,'CPP',''),
('CPV-OO',"Caprivi Strip",None,'CPV',''),
('CRL-MF',"Micronesia Federated States",None,'CRL',''),
('CRL-PA',"Palau",'PW','CRL',''),
('CRZ-OO',"Crozet Is.",None,'CRZ',''),
('CTA-OO',"Chita",None,'CTA',''),
('CTM-OO',"Chatham Is.",None,'CTM',''),
('CUB-OO',"Cuba",'CU','CUB',''),
('CVI-OO',"Cape Verde",'CV','CVI',''),
('CYP-OO',"Cyprus",'CY','CYP',''),
('CZE-CZ',"Czech Republic",'CZ','CZE',''),
('CZE-SK',"Slovakia",'SK','CZE',''),
('DEL-OO',"Delaware",None,'DEL',''),
('DEN-OO',"Denmark",'DK','DEN',''),
('DJI-OO',"Djibouti",'DJ','DJI',''),
('DOM-OO',"Dominican Republic",'DO','DOM',''),
('DSV-OO',"Desventurados Is.",None,'DSV',''),
('EAI-OO',"East Aegean Is.",None,'EAI',''),
('EAS-OO',"Easter Is.",None,'EAS',''),
('ECU-OO',"Ecuador",'EC','ECU',''),
('EGY-OO',"Egypt",'EG','EGY',''),
('EHM-AP',"Arunachal Pradesh",None,'EHM',''),
('EHM-BH',"Bhutan",'BT','EHM',''),
('EHM-DJ',"Darjiling",None,'EHM',''),
('EHM-SI',"Sikkim",None,'EHM',''),
('ELS-OO',"El Salvador",'SV','ELS',''),
('EQG-OO',"Equatorial Guinea",'GQ','EQG',''),
('ERI-OO',"Eritrea",'ER','ERI',''),
('ETH-OO',"Ethiopia",'ET','ETH',''),
('FAL-OO',"Falkland Is.",None,'FAL',''),
('FIJ-OO',"Fiji",'FJ','FIJ',''),
('FIN-OO',"Finland",'FI','FIN',''),
('FLA-OO',"Florida",None,'FLA',''),
('FOR-OO',"Føroyar",None,'FOR',''),
('FRA-CI',"Channel Is.",None,'FRA',''),
('FRA-FR',"France",'FR','FRA',''),
('FRA-MO',"Monaco",'MC','FRA',''),
('FRG-OO',"French Guiana",'GF','FRG',''),
('GAB-OO',"Gabon",'GA','GAB',''),
('GAL-OO',"Galápagos",None,'GAL',''),
('GAM-OO',"Gambia, The",None,'GAM',''),
('GEO-OO',"Georgia",'GE','GEO',''),
('GER-OO',"Germany",'DE','GER',''),
('GGI-AN',"Annobón",None,'GGI',''),
('GGI-BI',"Bioko",None,'GGI',''),
('GGI-PR',"Principe",None,'GGI',''),
('GGI-ST',"São Tomé",None,'GGI',''),
('GHA-OO',"Ghana",'GH','GHA',''),
('GIL-OO',"Gilbert Is.",None,'GIL',''),
('GNB-OO',"Guinea-Bissau",'GW','GNB',''),
('GNL-OO',"Greenland",'GL','GNL',''),
('GRB-OO',"Great Britain",None,'GRB',''),
('GRC-OO',"Greece",'GR','GRC',''),
('GST-BA',"Bahrain",'BH','GST',''),
('GST-QA',"Qatar",'QA','GST',''),
('GST-UA',"United Arab Emirates",'AE','GST',''),
('GUA-OO',"Guatemala",'GT','GUA',''),
('GUI-OO',"Guinea",'GN','GUI',''),
('GUY-OO',"Guyana",'GY','GUY',''),
('HAI-HA',"Haiti",'HT','HAI',''),
('HAI-NI',"Navassa I.",None,'HAI',''),
('HAW-HI',"Hawaiian Is.",None,'HAW',''),
('HAW-JI',"Johnston I.",None,'HAW',''),
('HAW-MI',"Midway Is.",None,'HAW',''),
('HBI-OO',"Howland-Baker Is.",None,'HBI',''),
('HMD-OO',"Heard-McDonald Is.",None,'HMD',''),
('HON-OO',"Honduras",'HN','HON',''),
('HUN-OO',"Hungary",'HU','HUN',''),
('ICE-OO',"Iceland",'IS','ICE',''),
('IDA-OO',"Idaho",None,'IDA',''),
('ILL-OO',"Illinois",None,'ILL',''),
('IND-AP',"Andhra Pradesh",None,'IND',''),
('IND-BI',"Bihar",None,'IND',''),
('IND-CH',"Chandigarh",None,'IND',''),
('IND-DD',"Dadra-Nagar-Haveli",None,'IND',''),
('IND-DE',"Delhi",None,'IND',''),
('IND-DI',"Diu",None,'IND',''),
('IND-DM',"Daman",None,'IND',''),
('IND-GO',"Goa",None,'IND',''),
('IND-GU',"Gujarat",None,'IND',''),
('IND-HA',"Haryana",None,'IND',''),
('IND-KE',"Kerala",None,'IND',''),
('IND-KL',"Karaikal",None,'IND',''),
('IND-KT',"Karnataka",None,'IND',''),
('IND-MH',"Mahé",None,'IND',''),
('IND-MP',"Madhya Pradesh",None,'IND',''),
('IND-MR',"Maharashtra",None,'IND',''),
('IND-OR',"Orissa",None,'IND',''),
('IND-PO',"Pondicherry",None,'IND',''),
('IND-PU',"Punjab",None,'IND',''),
('IND-RA',"Rajasthan",None,'IND',''),
('IND-TN',"Tamil Nadu",None,'IND',''),
('IND-UP',"Uttar Pradesh",None,'IND',''),
('IND-WB',"West Bengal",None,'IND',''),
('IND-YA',"Yanam",None,'IND',''),
('INI-OO',"Indiana",None,'INI',''),
('IOW-OO',"Iowa",None,'IOW',''),
('IRE-IR',"Ireland",'IE','IRE',''),
('IRE-NI',"Northern Ireland",None,'IRE',''),
('IRK-OO',"Irkutsk",None,'IRK',''),
('IRN-OO',"Iran",None,'IRN',''),
('IRQ-OO',"Iraq",'IQ','IRQ',''),
('ITA-IT',"Italy",'IT','ITA',''),
('ITA-SM',"San Marino",'SM','ITA',''),
('ITA-VC',"Vatican City",None,'ITA',''),
('IVO-OO',"Ivory Coast",None,'IVO',''),
('JAM-OO',"Jamaica",'JM','JAM',''),
('JAP-HK',"Hokkaido",None,'JAP',''),
('JAP-HN',"Honshu",None,'JAP',''),
('JAP-KY',"Kyushu",None,'JAP',''),
('JAP-SH',"Shikoku",None,'JAP',''),
('JAW-OO',"Jawa",None,'JAW',''),
('JNF-OO',"Juan Fernández Is.",None,'JNF',''),
('KAM-OO',"Kamchatka",None,'KAM',''),
('KAN-OO',"Kansas",None,'KAN',''),
('KAZ-OO',"Kazakhstan",'KZ','KAZ',''),
('KEG-OO',"Kerguelen",None,'KEG',''),
('KEN-OO',"Kenya",'KE','KEN',''),
('KER-OO',"Kermadec Is.",None,'KER',''),
('KGZ-OO',"Kirgizistan",None,'KGZ',''),
('KHA-OO',"Khabarovsk",None,'KHA',''),
('KOR-NK',"North Korea",None,'KOR',''),
('KOR-SK',"South Korea",None,'KOR',''),
('KRA-OO',"Krasnoyarsk",None,'KRA',''),
('KRI-OO',"Kriti",None,'KRI',''),
('KRY-OO',"Krym",None,'KRY',''),
('KTY-OO',"Kentucky",None,'KTY',''),
('KUR-OO',"Kuril Is.",None,'KUR',''),
('KUW-OO',"Kuwait",'KW','KUW',''),
('KZN-OO',"Kazan-retto",None,'KZN',''),
('LAB-OO',"Labrador",None,'LAB',''),
('LAO-OO',"Laos",None,'LAO',''),
('LBR-OO',"Liberia",'LR','LBR',''),
('LBS-LB',"Lebanon",'LB','LBS',''),
('LBS-SY',"Syria",None,'LBS',''),
('LBY-OO',"Libya",None,'LBY',''),
('LDV-OO',"Laccadive Is.",None,'LDV',''),
('LEE-AB',"Antigua-Barbuda",None,'LEE',''),
('LEE-AG',"Anguilla",'AI','LEE',''),
('LEE-AV',"Aves I.",None,'LEE',''),
('LEE-BV',"British Virgin Is.",None,'LEE',''),
('LEE-GU',"Guadeloupe",'GP','LEE',''),
('LEE-MO',"Montserrat",'MS','LEE',''),
('LEE-NL',"Netherlands Leeward Is.",None,'LEE',''),
('LEE-SK',"St.Kitts-Nevis",None,'LEE',''),
('LEE-SM',"St.Martin-St.Barthélémy",None,'LEE',''),
('LEE-VI',"Virgin Is.",None,'LEE',''),
('LES-OO',"Lesotho",'LS','LES',''),
('LIN-KI',"Kiribati Line Is.",None,'LIN',''),
('LIN-US',"U.S. Line Is.",None,'LIN',''),
('LOU-OO',"Louisiana",None,'LOU',''),
('LSI-BA',"Bali",None,'LSI',''),
('LSI-ET',"East Timor",None,'LSI',''),
('LSI-LS',"Lesser Sunda Is.",None,'LSI',''),
('MAG-OO',"Magadan",None,'MAG',''),
('MAI-OO',"Maine",None,'MAI',''),
('MAN-OO',"Manitoba",None,'MAN',''),
('MAQ-OO',"Macquarie Is.",None,'MAQ',''),
('MAS-OO',"Masachusettes",None,'MAS',''),
('MAU-OO',"Mauritius",'MU','MAU',''),
('MCI-OO',"Mozambique Channel Is.",None,'MCI',''),
('MCS-OO',"Marcus I.",None,'MCS',''),
('MDG-OO',"Madagascar",'MG','MDG',''),
('MDR-OO',"Madeira",None,'MDR',''),
('MDV-OO',"Maldives",'MV','MDV',''),
('MIC-OO',"Michigan",None,'MIC',''),
('MIN-OO',"Minnesota",None,'MIN',''),
('MLI-OO',"Mali",'ML','MLI',''),
('MLW-OO',"Malawi",'MW','MLW',''),
('MLY-PM',"Peninsular Malaysia",None,'MLY',''),
('MLY-SI',"Singapore",'SG','MLY',''),
('MNT-OO',"Montana",None,'MNT',''),
('MOL-OO',"Maluku",None,'MOL',''),
('MON-OO',"Mongolia",'MN','MON',''),
('MOR-MO',"Morocco",'MA','MOR',''),
('MOR-SP',"Spanish North African Territories",None,'MOR',''),
('MOZ-OO',"Mozambique",'MZ','MOZ',''),
('MPE-OO',"Marion-Prince Edward Is.",None,'MPE',''),
('MRN-GU',"Guam",'GU','MRN',''),
('MRN-NM',"Northern Marianas",None,'MRN',''),
('MRQ-OO',"Marquesas",None,'MRQ',''),
('MRS-OO',"Marshall Is.",None,'MRS',''),
('MRY-OO',"Maryland",None,'MRY',''),
('MSI-OO',"Mississippi",None,'MSI',''),
('MSO-OO',"Missouri",None,'MSO',''),
('MTN-OO',"Mauritania",'MR','MTN',''),
('MXC-DF',"Mexico Distrito Federal",None,'MXC',''),
('MXC-ME',"México State",None,'MXC',''),
('MXC-MO',"Morelos",None,'MXC',''),
('MXC-PU',"Puebla",None,'MXC',''),
('MXC-TL',"Tlaxcala",None,'MXC',''),
('MXE-AG',"Aguascalientes",None,'MXE',''),
('MXE-CO',"Coahuila",None,'MXE',''),
('MXE-CU',"Chihuahua",None,'MXE',''),
('MXE-DU',"Durango",None,'MXE',''),
('MXE-GU',"Guanajuato",None,'MXE',''),
('MXE-HI',"Hidalgo",None,'MXE',''),
('MXE-NL',"Nuevo León",None,'MXE',''),
('MXE-QU',"Querétaro",None,'MXE',''),
('MXE-SL',"San Luis Potosí",None,'MXE',''),
('MXE-TA',"Tamaulipas",None,'MXE',''),
('MXE-ZA',"Zacatecas",None,'MXE',''),
('MXG-VC',"Veracruz",None,'MXG',''),
('MXI-GU',"Guadalupe I.",None,'MXI',''),
('MXI-RA',"Rocas Alijos",None,'MXI',''),
('MXI-RG',"Revillagigedo Is.",None,'MXI',''),
('MXN-BC',"Baja California",None,'MXN',''),
('MXN-BS',"Baja California Sur",None,'MXN',''),
('MXN-SI',"Sinaloa",None,'MXN',''),
('MXN-SO',"Sonora",None,'MXN',''),
('MXS-CL',"Colima",None,'MXS',''),
('MXS-GR',"Guerrero",None,'MXS',''),
('MXS-JA',"Jalisco",None,'MXS',''),
('MXS-MI',"Michoacán",None,'MXS',''),
('MXS-NA',"Nayarit",None,'MXS',''),
('MXS-OA',"Oaxaca",None,'MXS',''),
('MXT-CA',"Campeche",None,'MXT',''),
('MXT-CI',"Chiapas",None,'MXT',''),
('MXT-QR',"Quintana Roo",None,'MXT',''),
('MXT-TB',"Tabasco",None,'MXT',''),
('MXT-YU',"Yucatán",None,'MXT',''),
('MYA-OO',"Myanmar",'MM','MYA',''),
('NAM-OO',"Namibia",'NA','NAM',''),
('NAT-OO',"KwaZulu-Natal",None,'NAT',''),
('NBR-OO',"New Brunswick",None,'NBR',''),
('NCA-OO',"North Carolina",None,'NCA',''),
('NCB-OO',"Nicobar Is.",None,'NCB',''),
('NCS-DA',"Dagestan",None,'NCS',''),
('NCS-KB',"Kabardino-Balkariya",None,'NCS',''),
('NCS-KC',"Karacheyevo-Cherkessiya",None,'NCS',''),
('NCS-KR',"Krasnodar",None,'NCS',''),
('NCS-SO',"Severo-Osetiya",None,'NCS',''),
('NCS-ST',"Stavropol",None,'NCS',''),
('NDA-OO',"North Dakota",None,'NDA',''),
('NEB-OO',"Nebraska",None,'NEB',''),
('NEP-OO',"Nepal",'NP','NEP',''),
('NET-OO',"Netherlands",'NL','NET',''),
('NEV-OO',"Nevada",None,'NEV',''),
('NFK-LH',"Lord Howe I.",None,'NFK',''),
('NFK-NI',"Norfolk I.",None,'NFK',''),
('NFL-NE',"Newfoundland",None,'NFL',''),
('NFL-SP',"St.Pierre-Miquelon",None,'NFL',''),
('NGA-OO',"Nigeria",'NG','NGA',''),
('NGR-OO',"Niger",'NE','NGR',''),
('NIC-OO',"Nicaragua",'NI','NIC',''),
('NLA-BO',"Bonaire",None,'NLA',''),
('NLA-CU',"Curaçao",None,'NLA',''),
('NNS-OO',"Nansei-shoto",None,'NNS',''),
('NOR-OO',"Norway",'NO','NOR',''),
('NRU-OO',"Nauru",'NR','NRU',''),
('NSC-OO',"Nova Scotia",None,'NSC',''),
('NSW-CT',"Australian Capital Territory",None,'NSW',''),
('NSW-NS',"New South Wales",None,'NSW',''),
('NTA-OO',"Northern Territory",None,'NTA',''),
('NUE-OO',"Niue",'NU','NUE',''),
('NUN-OO',"Nunavut",None,'NUN',''),
('NWC-OO',"New Caledonia",'NC','NWC',''),
('NWG-IJ',"Irian Jaya",None,'NWG',''),
('NWG-PN',"Papua New Guinea",'PG','NWG',''),
('NWH-OO',"New Hampshire",None,'NWH',''),
('NWJ-OO',"New Jersey",None,'NWJ',''),
('NWM-OO',"New Mexico",None,'NWM',''),
('NWT-OO',"Northwest Territories",None,'NWT',''),
('NWY-OO',"New York",None,'NWY',''),
('NZN-OO',"New Zealand North",None,'NZN',''),
('NZS-OO',"New Zealand South",None,'NZS',''),
('OFS-OO',"Free State",None,'OFS',''),
('OGA-OO',"Ogasawara-shoto",None,'OGA',''),
('OHI-OO',"Ohio",None,'OHI',''),
('OKL-OO',"Oklahoma",None,'OKL',''),
('OMA-OO',"Oman",'OM','OMA',''),
('ONT-OO',"Ontario",None,'ONT',''),
('ORE-OO',"Oregon",None,'ORE',''),
('PAK-OO',"Pakistan",'PK','PAK',''),
('PAL-IS',"Israel",'IL','PAL',''),
('PAL-JO',"Jordan",'JO','PAL',''),
('PAN-OO',"Panamá",None,'PAN',''),
('PAR-OO',"Paraguay",'PY','PAR',''),
('PEI-OO',"Prince Edward I.",None,'PEI',''),
('PEN-OO',"Pennsylvania",None,'PEN',''),
('PER-OO',"Peru",'PE','PER',''),
('PHI-OO',"Philippines",'PH','PHI',''),
('PHX-OO',"Phoenix Is.",None,'PHX',''),
('PIT-OO',"Pitcairn Is.",None,'PIT',''),
('POL-OO',"Poland",'PL','POL',''),
('POR-OO',"Portugal",'PT','POR',''),
('PRM-OO',"Primorye",None,'PRM',''),
('PUE-OO',"Puerto Rico",'PR','PUE',''),
('QLD-CS',"Coral Sea Is. Territory",None,'QLD',''),
('QLD-QU',"Queensland",None,'QLD',''),
('QUE-OO',"Québec",None,'QUE',''),
('REU-OO',"Réunion",'RE','REU',''),
('RHO-OO',"Rhode I.",None,'RHO',''),
('ROD-OO',"Rodrigues",None,'ROD',''),
('ROM-OO',"Romania",'RO','ROM',''),
('RUC-OO',"Central European Russia",None,'RUC',''),
('RUE-OO',"East European Russia",None,'RUE',''),
('RUN-OO',"North European Russia",None,'RUN',''),
('RUS-OO',"South European Russia",None,'RUS',''),
('RUW-OO',"Northwest European Russia",None,'RUW',''),
('RWA-OO',"Rwanda",'RW','RWA',''),
('SAK-OO',"Sakhalin",None,'SAK',''),
('SAM-AS',"American Samoa",'AS','SAM',''),
('SAM-WS',"Samoa",'WS','SAM','Western Samoa'),
('SAR-OO',"Sardegna",None,'SAR',''),
('SAS-OO',"Saskatchewan",None,'SAS',''),
('SAU-OO',"Saudi Arabia",'SA','SAU',''),
('SCA-OO',"South Carolina",None,'SCA',''),
('SCI-OO',"Society Is.",None,'SCI',''),
('SCS-PI',"Paracel Is.",None,'SCS',''),
('SCS-SI',"Spratly Is.",None,'SCS',''),
('SCZ-OO',"Santa Cruz Is.",None,'SCZ',''),
('SDA-OO',"South Dakota",None,'SDA',''),
('SEL-OO',"Selvagens",None,'SEL',''),
('SEN-OO',"Senegal",'SN','SEN',''),
('SEY-OO',"Seychelles",'SC','SEY',''),
('SGE-OO',"South Georgia",None,'SGE',''),
('SIC-MA',"Malta",'MT','SIC',''),
('SIC-SI',"Sicilia",None,'SIC',''),
('SIE-OO',"Sierra Leone",'SL','SIE',''),
('SIN-OO',"Sinai",None,'SIN',''),
('SOA-OO',"South Australia",None,'SOA',''),
('SOC-OO',"Socotra",None,'SOC',''),
('SOL-NO',"North Solomons",None,'SOL',''),
('SOL-SO',"South Solomons",None,'SOL',''),
('SOM-OO',"Somalia",'SO','SOM',''),
('SPA-AN',"Andorra",'AD','SPA',''),
('SPA-GI',"Gilbraltar",None,'SPA',''),
('SPA-SP',"Spain",'ES','SPA',''),
('SRL-OO',"Sri Lanka",'LK','SRL',''),
('SSA-OO',"South Sandwich Is.",None,'SSA',''),
('STH-OO',"St.Helena",None,'STH',''),
('SUD-OO',"Sudan",'SD','SUD',''),
('SUL-OO',"Sulawesi",None,'SUL',''),
('SUM-OO',"Sumatera",None,'SUM',''),
('SUR-OO',"Surinam",None,'SUR',''),
('SVA-OO',"Svalbard",None,'SVA',''),
('SWC-CC',"Colombian Caribbean Is.",None,'SWC',''),
('SWC-HC',"Honduran Caribbean Is.",None,'SWC',''),
('SWC-NC',"Nicaraguan Caribbean Is.",None,'SWC',''),
('SWE-OO',"Sweden",'SE','SWE',''),
('SWI-OO',"Switzerland",'CH','SWI',''),
('SWZ-OO',"Swaziland",'SZ','SWZ',''),
('TAI-OO',"Taiwan",None,'TAI',''),
('TAN-OO',"Tanzania",None,'TAN',''),
('TAS-OO',"Tasmania",None,'TAS',''),
('TCI-OO',"Turks-Caicos Is.",None,'TCI',''),
('TCS-AB',"Abkhaziya",None,'TCS',''),
('TCS-AD',"Adzhariya",None,'TCS',''),
('TCS-AR',"Armenia",'AM','TCS',''),
('TCS-AZ',"Azerbaijan",'AZ','TCS',''),
('TCS-GR',"Georgia",'GE','TCS',''),
('TCS-NA',"Nakhichevan",None,'TCS',''),
('TCS-NK',"Nagorno Karabakh",None,'TCS',''),
('TDC-OO',"Tristan da Cunha",None,'TDC',''),
('TEN-OO',"Tennessee",None,'TEN',''),
('TEX-OO',"Texas",None,'TEX',''),
('THA-OO',"Thailand",'TH','THA',''),
('TKM-OO',"Turkmenistan",'TM','TKM',''),
('TOG-OO',"Togo",'TG','TOG',''),
('TOK-MA',"Manihiki Is.",None,'TOK',''),
('TOK-SW',"Swains I.",None,'TOK',''),
('TOK-TO',"Tokelau",'TK','TOK',''),
('TON-OO',"Tonga",'TO','TON',''),
('TRT-OO',"Trinidad-Tobago",None,'TRT',''),
('TUA-OO',"Tuamotu",None,'TUA',''),
('TUB-OO',"Tubuai Is.",None,'TUB',''),
('TUE-OO',"Turkey-in-Europe",None,'TUE',''),
('TUN-OO',"Tunisia",'TN','TUN',''),
('TUR-OO',"Turkey",'TR','TUR',''),
('TUV-OO',"Tuvalu",'TV','TUV',''),
('TVA-OO',"Tuva",None,'TVA',''),
('TVL-GA',"Gauteng",None,'TVL',''),
('TVL-MP',"Mpumalanga",None,'TVL',''),
('TVL-NP',"Northern Province",None,'TVL',''),
('TVL-NW',"North-West Province",None,'TVL',''),
('TZK-OO',"Tadzhikistan",None,'TZK',''),
('UGA-OO',"Uganda",'UG','UGA',''),
('UKR-MO',"Moldova",None,'UKR',''),
('UKR-UK',"Ukraine",'UA','UKR',''),
('URU-OO',"Uruguay",'UY','URU',''),
('UTA-OO',"Utah",None,'UTA',''),
('UZB-OO',"Uzbekistan",'UZ','UZB',''),
('VAN-OO',"Vanuatu",'VU','VAN',''),
('VEN-OO',"Venezuela",None,'VEN',''),
('VER-OO',"Vermont",None,'VER',''),
('VIC-OO',"Victoria",None,'VIC',''),
('VIE-OO',"Vietnam",None,'VIE',''),
('VNA-OO',"Venezuelan Antilles",None,'VNA',''),
('VRG-OO',"Virginia",None,'VRG',''),
('WAK-OO',"Wake I.",None,'WAK',''),
('WAL-OO',"Wallis-Futuna Is.",None,'WAL',''),
('WAS-OO',"Washington",None,'WAS',''),
('WAU-AC',"Ashmore-Cartier Is.",None,'WAU',''),
('WAU-WA',"Western Australia",None,'WAU',''),
('WDC-OO',"District of Columbia",None,'WDC',''),
('WHM-HP',"Himachal Pradesh",None,'WHM',''),
('WHM-JK',"Jammu-Kashmir",None,'WHM',''),
('WHM-UT',"Uttaranchal",None,'WHM',''),
('WIN-BA',"Barbados",'BB','WIN',''),
('WIN-DO',"Dominica",'DM','WIN',''),
('WIN-GR',"Grenada",'GD','WIN',''),
('WIN-MA',"Martinique",'MQ','WIN',''),
('WIN-SL',"St.Lucia",None,'WIN',''),
('WIN-SV',"St.Vincent",None,'WIN',''),
('WIS-OO',"Wisconsin",None,'WIS',''),
('WSA-OO',"Western Sahara",'EH','WSA',''),
('WSB-OO',"West Siberia",None,'WSB',''),
('WVA-OO',"West Virginia",None,'WVA',''),
('WYO-OO',"Wyoming",None,'WYO',''),
('XMS-OO',"Christmas I.",None,'XMS',''),
('YAK-OO',"Yakutskiya",None,'YAK',''),
('YEM-NY',"North Yemen",None,'YEM',''),
('YEM-SY',"South Yemen",None,'YEM',''),
('YUG-BH',"Bosnia-Herzegovina",None,'YUG',''),
('YUG-CR',"Croatia",'HR','YUG',''),
('YUG-KO',"Kosovo",None,'YUG',''),
('YUG-MA',"Macedonia",None,'YUG',''),
('YUG-MN',"Montenegro",'ME','YUG',''),
('YUG-SE',"Serbia",'RS','YUG',''),
('YUG-SL',"Slovenia",'SI','YUG',''),
('YUK-OO',"Yukon",None,'YUK',''),
('ZAI-OO',"Zaire",None,'ZAI',''),
('ZAM-OO',"Zambia",'ZM','ZAM',''),
('ZIM-OO',"Zimbabwe",'ZW','ZIM',''),
('IND-JK',"Jharkhand",None,'IND',''),
('IND-CT',"Chattisgarh",None,'IND',''),
('NCS-CH',"Chechnya",None,'NCS','Una de las partes en que se divide Checheno-Ingushetiya'),
('NCS-IN',"Ingushetiya",None,'NCS','Una de las partes en que se divide Checheno-Ingushetiya'),
('CHC-CQ',"Chongqing",None,'CHC','Anteriormente parte de Sichuan'),
]

extra_names = [("Kyrgyzstan","Kirgizistan"),
("Russia", "Russian Federation"),
("Macao", "Macau"),
("Panama", "Panamá"),
("Burkina Faso", "Burkina"),
("Gambia", "Gambia, The"),
("Côte d'Ivoire", "Ivory Coast"),
("Trinidad and Tobago", "Trinidad-Tobago"),
("Antigua and Barbuda", "Antigua-Barbuda"),
("Tajikstan", "Tadzhikistan"),
("Democratic Republic of the Congo", "Zaire"),
("Transvaal", "Northern Provinces"),
("Burma", "Myanmar"),
("Yukon Territory", "Yukon"),
("South America", "Southern America"),
("Yamalo-Nenets", "Yamalia"),
("Central African Rep.", "Central African Republic"),
]

extra_names_tdwg = [("Turkey in Asia", "TUR"),
("Turkey in Europe", "TUE"),
("Yakutiya", "YAK"),
("Yakutia", "YAK"),
("Kirgizstan", "KGZ"),
("Natal", "NAT"),
("Gruzia", "TCS-GR")
]

extra_ISO = {"VE": "Venezuela",
"TZ": "Tanzania",
"TW": "Taiwan",
"LY": "Libya",
"IR": "Iran",
"KR": "South Korea",
"KP": "North Korea",
"LA": "Laos",
"FK": "Falkland Is.",
"BO": "Bolivia",
"SJ": "Svalbard",
"BA": "Bosnia-Herzegovina",
"MK": "Macedonia",
"SY": "Syria",
"MD": "Moldova",
"BN": "Brunei",
"VN": "Vietnam",
"KG": "Kirgizistan",
"MO": "Macau",
"PA": "Panamá",
"BF": "Burkina",
"GM": "Gambia, The",
"CI": "Ivory Coast",
"TT": "Trinidad-Tobago",
"AG": "Antigua-Barbuda",
"TJ": "Tadzhikistan",
"CD": "Zaire",
"GB": "United Kingdom"}

extra_countries = [\
("Russian Federation", None, ["RUC","RUE","RUN","RUS","RUW","NCS","30","31"]),
("Turkey", None, ["TUE","TUR"]),
("India", "40", ["IND","WHM","EHM-SI","EHM-DJ","EHM-AP","ASS","LDV"]),
("United Kingdom", "10", ["GRB","IRE-NI"]),
# The US here is only the "lower 48"--excluding Alaska and Hawai'i
("United States", "7", ["78","75","74","77","76","73"]),
("Canada", "7", ["71","72","NWT","YUK","NUN"]),
("Chile", "85", ["CLC","CLN","CLS"]),
("Argentina", "85", ["AGE","AGW","AGS"]),
("South Africa", "27", ["TVL","CPP","OFS","NAT"]),
("Malaysia", "42",  ["MLY-PM","BOR-SB","BOR-SR"]),
("Indonesia", "4", ["LSI-BA","LSI-LS","BOR-KA","SUL","JAW",\
                            "SUM","MOL","NWG-IJ"]),
("Yamalia", "WSB",[]),
]

tdwg_by_level = [None, {'1', '3', '2', '5', '4', '7', '6', '9', '8'}, {'42', '43', '60', '61', '62', '63', '82', '83', '80', '81', '84', '85', '24', '25', '26', '27', '20', '21', '22', '23', '28', '29', '40', '41', '51', '77', '76', '75', '38', '73', '72', '71', '70', '91', '90', '79', '78', '11', '10', '13', '12', '14', '33', '32', '31', '30', '37', '36', '35', '34', '74', '50'}, {'CPI', 'MNT', 'AGE', 'SIC', 'YUK', 'SIE', 'CPP', 'AGS', 'SIN', 'CPV', 'TAS', 'SPA', 'TAN', 'TAI', 'VAN', 'SCZ', 'DJI', 'GIL', 'FIN', 'MAQ', 'MAS', 'LAO', 'MAU', 'NWH', 'LAB', 'JNF', 'MAG', 'ALG', 'MAI', 'MAN', 'CMN', 'GUA', 'BEN', 'GUI', 'AGW', 'BER', 'GUY', 'HUN', 'LSI', 'YUG', 'BLZ', 'DSV', 'BLT', 'NFL', 'NFK', 'BLR', 'GRB', 'GRC', 'RWA', 'ALU', 'CBD', 'MXT', 'MXS', 'MXN', 'INI', 'MXI', 'IND', 'MXG', 'MXE', 'MXC', 'NOR', 'BZS', 'DOM', 'REU', 'BZE', 'BZC', 'CTA', 'CTM', 'BZL', 'WHM', 'MDG', 'MDR', 'GGI', 'VER', 'AUT', 'VEN', 'PUE', 'ALD', 'ALB', 'ALA', 'QUE', 'NTA', 'ALT', 'ORE', 'CAY', 'LOU', 'UKR', 'TOK', 'TON', 'CAL', 'BAN', 'CAB', 'BAL', 'KOR', 'TOG', 'BAH', 'CAF', 'ERI', 'ONT', 'NZS', 'KTY', 'STH', 'ABT', 'NZN', 'NCS', 'NCA', 'QLD', 'OGA', 'EGY', 'MIN', 'MIC', 'FIJ', 'GHA', 'SAK', 'LBS', 'EHM', 'SAM', 'LBY', 'EQG', 'SAS', 'SAR', 'SAU', 'CHA', 'CHC', 'MRY', 'CHH', 'JAW', 'CHM', 'CHN', 'MRS', 'CHQ', 'MRN', 'CHS', 'CHT', 'CHX', 'ZIM', 'VIC', 'VIE', 'KAN', 'KAM', 'RHO', 'KAZ', 'NGA', 'NGR', 'UGA', 'TKM', 'NLA', 'HMD', 'RUC', 'AFG', 'SSA', 'CZE', 'SCA', 'ATP', 'WYO', 'ASS', 'ASC', 'ASP', 'SEY', 'NUN', 'ECU', 'ASK', 'SEN', 'SEL', 'CLM', 'CLN', 'FRG', 'CLC', 'YAK', 'CLS', 'KEN', 'OMA', 'NUE', 'KEG', 'TUR', 'TUV', 'ITA', 'TUN', 'TUA', 'TUB', 'TUE', 'SWI', 'SWC', 'SWE', 'FRA', 'SWZ', 'CYP', 'POR', 'POL', 'BIS', 'MLI', 'MLW', 'MLY', 'NCB', 'MCI', 'MCS', 'NRU', 'PAR', 'CKI', 'BGM', 'PAK', 'PAL', 'PAN', 'VNA', 'TZK', 'LIN', 'HAI', 'OFS', 'THA', 'HAW', 'PHI', 'ZAM', 'YEM', 'ZAI', 'NIC', 'PHX', 'WIN', 'CRZ', 'CRL', 'GER', 'BUR', 'TCS', 'LDV', 'BUL', 'TCI', 'GEO', 'FOR', 'LBR', 'AND', 'MOZ', 'ANG', 'MOR', 'KER', 'ANT', 'PRM', 'MON', 'MOL', 'COS', 'COR', 'ILL', 'PER', 'PEN', 'PEI', 'ETH', 'COM', 'COL', 'COO', 'CON', 'TVL', 'VRG', 'TVA', 'KRY', 'WAK', 'NDA', 'SRL', 'KRI', 'WAU', 'KRA', 'WAS', 'XMS', 'RUN', 'TDC', 'PIT', 'RUE', 'RUS', 'RUW', 'SCS', 'WSA', 'WSB', 'GNB', 'CVI', 'GNL', 'SOA', 'SOC', 'NSC', 'SOM', 'SOL', 'GAM', 'GAL', 'FAL', 'CHI', 'GAB', 'NSW', 'UTA', 'IRQ', 'MRQ', 'IRK', 'MTN', 'JAP', 'IRN', 'IRE', 'ARU', 'URU', 'OKL', 'JAM', 'SDA', 'ARI', 'ARK', 'WDC', 'HBI', 'GST', 'BOR', 'BOT', 'BOU', 'NAM', 'BOL', 'NAT', 'OHI', 'IOW', 'MYA', 'KHA', 'NNS', 'SVA', 'DEL', 'DEN', 'BZN', 'FLA', 'NWT', 'SCI', 'ROM', 'ROD', 'NWY', 'NWG', 'NWC', 'NWM', 'WVA', 'AMU', 'NWJ', 'TRT', 'MPE', 'IVO', 'SUD', 'KUW', 'SUM', 'SUL', 'NEB', 'KUR', 'SUR', 'NEV', 'NET', 'NEP', 'IDA', 'TEN', 'NBR', 'BKN', 'TEX', 'CGS', 'CUB', 'ELS', 'LEE', 'BRY', 'MDV', 'LES', 'WIS', 'KZN', 'AZO', 'MSO', 'WAL', 'MSI', 'EAI', 'UZB', 'EAS', 'SGE', 'CNT', 'HON', 'BRC', 'ICE', 'CNY', 'KGZ'}, {'ASS-NA', 'ALB-OO', 'ARK-OO', 'LEE-BV', 'ILL-OO', 'TCS-GR', 'MXT-CA', 'AGW-JU', 'BOL-OO', 'TOG-OO', 'CLS-MG', 'AFG-OO', 'MLY-SI', 'TOK-TO', 'EHM-BH', 'CPI-CL', 'CPI-CO', 'FIJ-OO', 'MAI-OO', 'AGS-NE', 'TAN-OO', 'NGA-OO', 'GAL-OO', 'ALG-OO', 'HAW-JI', 'SDA-OO', 'TDC-OO', 'NCS-CH', 'ELS-OO', 'UKR-MO', 'MAG-OO', 'CHM-LN', 'CLC-BI', 'WHM-UT', 'IND-JK', 'AGS-RN', 'SUR-OO', 'ECU-OO', 'SAS-OO', 'ARI-OO', 'CHN-GS', 'LES-OO', 'COO-OO', 'MXE-AG', 'BZN-PA', 'ASC-OO', 'BZS-SC', 'MXS-NA', 'SWC-HC', 'VAN-OO', 'BLT-ES', 'RWA-OO', 'CZE-CZ', 'IND-TN', 'SPA-AN', 'SEY-OO', 'IND-KT', 'CLN-TA', 'MXE-SL', 'AGW-CA', 'CMN-OO', 'ZIM-OO', 'CHS-GX', 'BLT-LI', 'CHS-GD', 'MCS-OO', 'NWG-IJ', 'BLT-LA', 'CHC-HU', 'IRK-OO', 'SIC-SI', 'TUN-OO', 'BZE-RN', 'OFS-OO', 'TOK-SW', 'SPA-SP', 'SCI-OO', 'TUB-OO', 'CHA-OO', 'CLS-AI', 'AGE-DF', 'AMU-OO', 'EAS-OO', 'NWT-OO', 'GST-QA', 'MRN-GU', 'IND-MH', 'SUM-OO', 'SAU-OO', 'FRA-CI', 'MXE-HI', 'RUN-OO', 'IND-MP', 'IND-MR', 'TVL-GA', 'WAL-OO', 'NIC-OO', 'VIE-OO', 'INI-OO', 'ASS-AS', 'YUG-SE', 'MXS-OA', 'FLA-OO', 'YUG-SL', 'ITA-VC', 'KUR-OO', 'CRZ-OO', 'AUT-LI', 'GHA-OO', 'KRY-OO', 'SCS-SI', 'CHM-HJ', 'MXE-ZA', 'LEE-SM', 'XMS-OO', 'LEE-SK', 'CPV-OO', 'CHS-ZJ', 'TCI-OO', 'BLT-KA', 'BLR-OO', 'VEN-OO', 'MXC-TL', 'MXS-JA', 'EHM-SI', 'FRG-OO', 'BZN-TO', 'LEE-AV', 'IND-PO', 'CNT-OO', 'CAF-OO', 'LEE-AB', 'COR-OO', 'LEE-AG', 'IND-PU', 'HAI-NI', 'STH-OO', 'KRA-OO', 'BZL-MG', 'ANT-OO', 'WAU-WA', 'LBS-LB', 'SUL-OO', 'IND-BI', 'IRE-NI', 'MXG-VC', 'KAZ-OO', 'EHM-AP', 'CLC-SA', 'DEN-OO', 'BZL-TR', 'MPE-OO', 'BAH-OO', 'BZN-AP', 'LAO-OO', 'BZN-AM', 'MRS-OO', 'BZN-AC', 'ARU-OO', 'BUL-OO', 'CLC-MA', 'BGM-BE', 'JAP-HN', 'IRQ-OO', 'JAP-HK', 'MOR-MO', 'OKL-OO', 'TUV-OO', 'UZB-OO', 'SAR-OO', 'ASS-TR', 'SAM-WS', 'MXS-MI', 'MXE-DU', 'URU-OO', 'WIN-DO', 'BAN-OO', 'KOR-SK', 'BZS-PR', 'JAM-OO', 'PUE-OO', 'PAR-OO', 'LSI-BA', 'YUG-BH', 'QUE-OO', 'MIN-OO', 'PEI-OO', 'IND-UP', 'TKM-OO', 'KAM-OO', 'MRY-OO', 'SIE-OO', 'TVA-OO', 'BOT-OO', 'YUG-KO', 'MXC-PU', 'TRT-OO', 'AGW-LR', 'CHS-FJ', 'NWJ-OO', 'UKR-UK', 'CHS-HK', 'CHC-YN', 'CHS-HN', 'CHS-HE', 'IDA-OO', 'PHI-OO', 'NET-OO', 'GIL-OO', 'NCS-IN', 'BUR-OO', 'OGA-OO', 'PEN-OO', 'AUT-AU', 'SOL-NO', 'PIT-OO', 'HBI-OO', 'CLS-LL', 'BZE-MA', 'MXE-GU', 'YEM-SY', 'CHQ-OO', 'RUS-OO', 'ALT-OO', 'PAL-JO', 'WIN-MA', 'AND-CO', 'NWY-OO', 'MOR-SP', 'IND-GU', 'CPP-WC', 'IND-GO', 'GEO-OO', 'BZC-GO', 'CAB-OO', 'HAW-MI', 'CHN-SA', 'MRQ-OO', 'HON-OO', 'AGS-SF', 'AGS-SC', 'CPP-EC', 'MLI-OO', 'NTA-OO', 'MSO-OO', 'EQG-OO', 'MTN-OO', 'LEE-VI', 'LBR-OO', 'CLN-AN', 'ICE-OO', 'SOL-SO', 'LOU-OO', 'KRI-OO', 'MXI-RG', 'YEM-NY', 'GAB-OO', 'ASK-OO', 'NSC-OO', 'CHS-KI', 'NWC-OO', 'ATP-OO', 'AGE-BA', 'LBY-OO', 'BZL-ES', 'ANG-OO', 'NNS-OO', 'NCB-OO', 'NZN-OO', 'CHC-SC', 'GST-BA', 'VER-OO', 'MXN-BS', 'SAM-AS', 'CVI-OO', 'MXE-NL', 'ONT-OO', 'JAW-OO', 'EHM-DJ', 'ALA-OO', 'RUW-OO', 'BZE-SE', 'NWM-OO', 'AGE-LP', 'AGE-ER', 'TUR-OO', 'JNF-OO', 'CYP-OO', 'YUK-OO', 'CLC-LA', 'PAK-OO', 'GUA-OO', 'BOR-SR', 'NSW-CT', 'BRY-OO', 'NLA-BO', 'GAM-OO', 'BOR-SB', 'BZE-AL', 'NUN-OO', 'CLM-OO', 'NFK-NI', 'MXE-CO', 'AGW-SA', 'SGE-OO', 'AGW-SE', 'MXE-CU', 'AGW-SJ', 'GUY-OO', 'AGW-SL', 'EAI-OO', 'KEN-OO', 'MAN-OO', 'NCS-SO', 'TOK-MA', 'WHM-JK', 'SCS-PI', 'NCS-ST', 'IVO-OO', 'HAI-HA', 'TZK-OO', 'YAK-OO', 'RUC-OO', 'GRB-OO', 'WIN-GR', 'THA-OO', 'CZE-SK', 'EGY-OO', 'BZC-MS', 'WIS-OO', 'BZC-MT', 'COM-MA', 'CBD-OO', 'MYA-OO', 'LAB-OO', 'NCA-OO', 'WSA-OO', 'NAT-OO', 'CKI-OO', 'AGE-FO', 'VIC-OO', 'NBR-OO', 'LDV-OO', 'TEX-OO', 'MDV-OO', 'SEL-OO', 'CHX-OO', 'BRC-OO', 'CRL-PA', 'MXS-GR', 'MOZ-OO', 'SOM-OO', 'MXT-TB', 'NUE-OO', 'IND-CH', 'IRE-IR', 'WDC-OO', 'IND-CT', 'FRA-MO', 'BZL-SP', 'MXN-SO', 'CON-OO', 'HAW-HI', 'CHT-OO', 'AGS-CB', 'NFL-NE', 'IND-DD', 'AGW-TU', 'LBS-SY', 'MRN-NM', 'IND-DM', 'NWG-PN', 'IND-DI', 'KUW-OO', 'POL-OO', 'AGS-TF', 'GGI-AN', 'KTY-OO', 'MSI-OO', 'TVL-NP', 'TVL-NW', 'TCS-AR', 'CLC-OH', 'WIN-BA', 'BZN-RO', 'BZN-RM', 'FAL-OO', 'ITA-IT', 'MXE-QU', 'MXT-CI', 'MXN-SI', 'YUG-CR', 'SEN-OO', 'LEE-GU', 'CGS-OO', 'SWC-CC', 'AGW-ME', 'KOR-NK', 'MON-OO', 'LEE-NL', 'COL-OO', 'ROM-OO', 'SOA-OO', 'AGE-MI', 'MXC-DF', 'COS-OO', 'NOR-OO', 'CHI-NM', 'BZE-PI', 'MCI-OO', 'WIN-SL', 'SIC-MA', 'BIS-OO', 'MDR-OO', 'BZE-PB', 'CHI-NX', 'BZE-PE', 'IND-YA', 'NEB-OO', 'CHN-HB', 'SOC-OO', 'PAN-OO', 'DEL-OO', 'TAS-OO', 'UGA-OO', 'SCZ-OO', 'LIN-US', 'NCS-KR', 'IND-DE', 'BOR-BR', 'NCS-DA', 'BZC-DF', 'CTM-OO', 'ETH-OO', 'SWC-NC', 'CHS-SH', 'AZO-OO', 'IND-OR', 'BZE-BA', 'MXC-MO', 'KER-OO', 'NSW-NS', 'MXC-ME', 'CLC-VA', 'CHN-SD', 'QLD-CS', 'MXS-CL', 'JAP-SH', 'MXT-QR', 'ALD-OO', 'PRM-OO', 'CLN-AT', 'SSA-OO', 'IND-AP', 'CHN-SX', 'GGI-ST', 'CHS-AH', 'PAL-IS', 'MXE-TA', 'CHM-JL', 'HMD-OO', 'YUG-MN', 'NGR-OO', 'MAU-OO', 'GNL-OO', 'WYO-OO', 'YUG-MA', 'ZAM-OO', 'ITA-SM', 'SVA-OO', 'LEE-MO', 'ASS-ME', 'SWZ-OO', 'CHS-JS', 'ASS-MA', 'DSV-OO', 'CHS-JX', 'ASS-MI', 'CUB-OO', 'VNA-OO', 'IND-RA', 'NCS-KB', 'NCS-KC', 'QLD-QU', 'KHA-OO', 'MOL-OO', 'NZS-OO', 'SPA-GI', 'AGE-CO', 'AGE-CN', 'AGE-CH', 'WVA-OO', 'COM-CO', 'ASP-OO', 'LIN-KI', 'NAM-OO', 'SIN-OO', 'BEN-OO', 'MAQ-OO', 'RHO-OO', 'BOU-OO', 'ALU-OO', 'NRU-OO', 'GUI-OO', 'CTA-OO', 'GNB-OO', 'CHN-TJ', 'ROD-OO', 'CHC-GZ', 'SWI-OO', 'TUA-OO', 'MDG-OO', 'NEP-OO', 'TON-OO', 'WIN-SV', 'AND-AN', 'GGI-PR', 'CNY-OO', 'DJI-OO', 'BZL-RJ', 'CLC-CO', 'POR-OO', 'CHH-OO', 'VRG-OO', 'NLA-CU', 'GRC-OO', 'WAS-OO', 'IND-KL', 'BGM-LU', 'BZE-FN', 'IND-KE', 'MXN-BC', 'TVL-MP', 'CPI-MA', 'MAS-OO', 'BOR-KA', 'ERI-OO', 'NEV-OO', 'BZS-RS', 'GER-OO', 'CAY-OO', 'CHS-MA', 'IND-WB', 'MIC-OO', 'UTA-OO', 'TAI-OO', 'CHS-MP', 'TUE-OO', 'SWE-OO', 'LSI-ET', 'CPP-NC', 'NFL-SP', 'BKN-OO', 'ORE-OO', 'NWH-OO', 'PER-OO', 'BAL-OO', 'WSB-OO', 'TEN-OO', 'SCA-OO', 'KGZ-OO', 'KAN-OO', 'KEG-OO', 'ABT-OO', 'ZAI-OO', 'KZN-OO', 'REU-OO', 'MLY-PM', 'BER-OO', 'CAL-OO', 'TCS-AZ', 'MXT-YU', 'BZE-CE', 'DOM-OO', 'IND-HA', 'TCS-AB', 'MXI-RA', 'CHC-CQ', 'TCS-AD', 'PHX-OO', 'IOW-OO', 'CRL-MF', 'NFK-LH', 'HUN-OO', 'MXI-GU', 'FRA-FR', 'WAK-OO', 'GST-UA', 'SAK-OO', 'MNT-OO', 'WHM-HP', 'NDA-OO', 'BLZ-OO', 'MLW-OO', 'SUD-OO', 'FIN-OO', 'RUE-OO', 'CHN-BJ', 'GGI-BI', 'WAU-AC', 'JAP-KY', 'IRN-OO', 'OMA-OO', 'LSI-LS', 'FOR-OO', 'TCS-NA', 'SRL-OO', 'TCS-NK', 'OHI-OO'}]



import csv
import os.path
import warnings
from taxonome import TaxonSet, Name, Taxon
from .base import UncertainSpeciesError

def _make_name(d, rawname, authfield):
    if authfield is True:
        return Name.from_string(rawname)
    
    if rawname.strip().endswith((" sp.", " spp.")):
        raise UncertainSpeciesError(rawname)
    if authfield:
        return Name(rawname, d.pop(authfield))
    else:
        return Name(rawname)


def iter_taxa(filehandle, namefield="Name", authfield="Author", progress=None,
                                                            *args, **kwargs):
    """Load taxa iteratively from a CSV file. It expects fields for name and author
    to be present, describing the taxa to use. Other fields will be stored as a
    dictionary in each taxon's .info property.
    
    If authfield is set to True, it will attempt to parse names in the namefield
    to names and authorities. If authfield is None, no authfield will be used.
    
    progress is an object used by the GUI to track progress. It is called
    periodically with the number of bytes read.
    
    Any other arguments are passed to the csv.DictReader object--see Python's
    documentation.
    """
    csvin = csv.DictReader(filehandle, *args, **kwargs)
    i = 0
    for line in csvin:
        rawname = line.pop(namefield)    # Raise KeyError if the field isn't there
        try:
            newtaxon = Taxon(_make_name(line, rawname, authfield))
        except Exception as e:
            warnings.warn("Failed to read taxon - error was: %s" % e)
            continue
        newtaxon.info = line
        yield newtaxon
        
        i += 1
        if progress and (i % 20 == 0):
            progress(filehandle.buffer.tell())

def load_taxa(filehandle, namefield="Name", authfield="Author", *args, **kwargs):
    """Load a TaxonSet from a CSV file. It expects fields for name and author
    to be present, describing the taxa to use. Other fields will be stored as a
    dictionary in each taxon's .info property.
    
    Any other arguments are passed to the csv.DictReader object--see Python's
    documentation.
    """
    taxa = TaxonSet()
    for tax in iter_taxa(filehandle, namefield, authfield, *args, **kwargs):
        taxa.add(tax)
    return taxa

def save_taxa(filehandle, taxonset, namefield="Name", authfield="Author", info_fields = [],
                write_distribution=False, *args, **kwargs):
    """Writes a TaxonSet to a CSV file. Accepted name and author will be
    recorded, along with any fields specified from .info.
    
    If info_fields is True, all fields will be saved. If write_distribution is
    None, distribution will be saved if any data is present. Both of these
    options need a collection of taxa which can be iterated over more than once.
    """
    if info_fields is True:
        info_fields = set()
        for t in taxonset:
            info_fields.update(t.info)
        info_fields = list(info_fields)
    fields = [namefield,authfield] + info_fields
    if (write_distribution is None) and any(t.distribution for t in taxonset):
        fields.append("Distribution")
        write_distribution = True
    csvout = csv.DictWriter(filehandle, fields, extrasaction='ignore', *args, **kwargs)
    csvout.writeheader()
    for ataxon in taxonset:
        towrite = {namefield:ataxon.name.plain, authfield:ataxon.name.authority}
        towrite.update(ataxon.info)
        towrite['Distribution'] = ataxon.distribution
        csvout.writerow(towrite)

def iter_individuals(filehandle, namefield="Name", authfield="Author", progress=None, **kwargs):
    """Read individuals from a CSV file - i.e. the same name may appear on more
    than one row.
    
    Yields pairs of name, info
    """
    csvin = csv.DictReader(filehandle, **kwargs)
    i = 0
    for line in csvin:
        rawname = line.pop(namefield)    # Raise KeyError if the field isn't there
        try:
            name = _make_name(line, rawname, authfield)
        except Exception as e:
            warnings.warn("Failed to read record - error was: %s" % e)
            continue
        yield name, line
        
        # Progress tracking
        i += 1
        if progress and (i % 20 == 0):
            progress(filehandle.buffer.tell())

def load_individuals(filehandle, namefield="Name", authfield="Author", **kwargs):
    """Load a TaxonSet representing the individuals in the CSV file. Each name
    may appear on several rows.
    """
    taxa = TaxonSet()
    for name, info in iter_individuals(filehandle, namefield, authfield, **kwargs):
        try:
            taxon = taxa[name]
        except KeyError:
            taxon = Taxon(name)
            taxa.add(taxon)
        taxon.individuals.append(info)
    return taxa

def save_individuals(filehandle, taxonset, namefield="Name", authfield="Author", **kwargs):
    """Save records of individuals (e.g. specimens, sightings) from a TaxonSet.
    
    This requires a collection which can be iterated more than once.
    """
    info_fields = set()
    for t in taxonset:
        for ind in t.individuals:
            info_fields.update(ind)
    fields = [namefield, authfield] + list(info_fields)
    
    csvout = csv.DictWriter(filehandle, fields, extrasaction='ignore', **kwargs)
    csvout.writeheader()
    
    for t in taxonset:
        name_fields = {namefield: t.name.plain, authfield:str(t.name.authority)}
        for ind in t.individuals:
            row = dict(ind)
            row.update(name_fields)
            csvout.writerow(row)

def load_synonyms(filehandle, synnamefield="Name", synauthfield="Author",\
                    accnamefield="Accname", accauthfield="Accauthor",\
                    get_info=True, *args, **kwargs):
    """Loads a set of synonyms from a CSV file. Returns a TaxonSet containing
    the taxa.
    
    *field parameters work like load_taxa.
    
    If get_info is True (the default), extra info from the lines with the
    accepted names is attached to the taxa.
    """
    holding = {}
    for synname, accname, info in iter_synonyms(filehandle, synnamefield, synauthfield,
                                accnamefield, accauthfield, *args, **kwargs):
        #print(repr(synname), repr(accname))
        if accname in holding:
            taxon = holding[accname]
        else:
            holding[accname] = taxon = Taxon(accname)
        
        # Add info from the line.
        if accname == synname:
            if get_info:
                taxon.info = info
        else:
            # Add synonym
            taxon.othernames.add(synname)
    
    taxonset = TaxonSet()
    for tax in holding.values():
        taxonset.add(tax)
    return taxonset
    
def iter_synonyms(filehandle, synnamefield="Name", synauthfield="Author",\
                    accnamefield="Accname", accauthfield="Accauthor",\
                    progress=None, *args, **kwargs):
    """Loads a set of synonyms from a CSV file. Yields triples of (synonym,
    accepted name, other fields dict).
    
    *field parameters work like load_taxa.
    """
    csvin = csv.DictReader(filehandle, *args, **kwargs)
    i = 0
    for line in csvin:
        rawaccname = line.pop(accnamefield)
        try:
            accname = _make_name(line, rawaccname, accauthfield)
        except Exception as e:
            warnings.warn("Failed to read accepted name - error was: %r" % e)
            continue
        rawsynname = line.pop(synnamefield)
        try:
            synname = _make_name(line, rawsynname, synauthfield)
        except Exception as e:
            warnings.warn("Failed to read synonym - error was: %r" % e)
            continue
        yield (synname, accname, line)
        
        i += 1
        if progress and (i % 20 == 0):
            progress(filehandle.buffer.tell())

def save_synonyms(filehandle, taxonset, snamefield="Name", sauthfield="Author",\
                    accnamefield="Accname", accauthfield="Accauthor",\
                    include_accepted_names=True, *args, **kwargs):
    """Writes a set of synonyms to a CSV file. (This only uses the synonyms in
    the .othernames property of each Taxon, not the names in the TaxonSet's
    index."""
    fields = [snamefield, sauthfield, accnamefield, accauthfield]
    csvout = csv.writer(filehandle, *args, **kwargs)
    csvout.writerow(fields)
    for taxon in taxonset:
        if include_accepted_names:
            csvout.writerow([taxon.name.plain, taxon.name.authority]*2)
        for synname in taxon.othernames:
            csvout.writerow([synname.plain, synname.authority,\
                        taxon.name.plain, taxon.name.authority])



"""
Graph ABC, written by Matteo Dell'Amico for Python 3.

http://www.linux.it/~della/GraphABC/

(MIT license obtained by e-mail)
"""
#Copyright (c) 2009 Matteo Dell'Amico

#Permission is hereby granted, free of charge, to any person obtaining a copy of
#this software and associated documentation files (the "Software"), to deal in
#the Software without restriction, including without limitation the rights to
#use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
#of the Software, and to permit persons to whom the Software is furnished to do
#so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

from abc import ABCMeta, abstractmethod
from collections import Set, Mapping, MutableSet, MutableMapping

class Graph(metaclass=ABCMeta):
    """Graph abstract base class.

    Edges are always labeled, with a default label of 1.

    The following invariants always apply:
    (n1, n2) in g.edges implies {n1, n2} <= g.nodes
    n not in g.nodes implies all(n not in edge for edge in g.edges)
    """

    class NodesView(Set):
        """ABC for the nodes."""

        def __init__(self, g):
            self.graph = g

    class EdgesView(Mapping):
        """ABC for the edges."""

        def __init__(self, g):
            self.graph = g

    def __init__(self):
        self.nodes = self.NodesView(self)
        self.edges = self.EdgesView(self)

    @abstractmethod
    def neighbors(self, node) -> Mapping:
        """Mapping from neighbors to label of corresponding edges.

        Do not try to modify the mapping resulting from this function:
        it may result in inconsistent behaviour.
        """

    @classmethod
    @abstractmethod
    def build(cls, edges, nodes=[]):
        """Returns a new graph with nodes and edges specified.

        If a node is present in edges but not in nodes, it will be added
        to nodes anyway.
        """

    @classmethod
    def copy_from(cls, g):
        """Returns a new graph with the same nodes and edges.

        An error is returned if a directed instance is called from an
        undirected one.
        """

        return cls.build(g.edges, g.nodes)

    def __eq__(self, other):
        return self.nodes == other.nodes and self.edges == other.edges

    def __ne__(self, other):
        return not (self == other)
    
class MutableGraph(Graph):

    @classmethod
    def from_graph(cls, g):
        res = cls()
        res.nodes |= g.nodes
        res.edges.update(g.edges)

    @classmethod
    def from_edges(cls, edges, nodes=[]):
        res = cls()
        res.nodes |= nodes
        res.edges.update(edges)

    class NodesView(MutableSet, Graph.NodesView):
        
        @abstractmethod
        def discard(self, node):
            """All edges incident to node should be discarded as well."""
    
    class EdgesView(MutableMapping, Graph.EdgesView):
        
        @abstractmethod
        def __setitem__(self, nodes, edge):
            """Nodes of the edge should be added to the graph's nodes."""

        # Convenience methods that add (default edge label: 1) and
        # discard edges mimicking a Set-like two-argument interface.

        def add(self, n1, n2):
            self[n1, n2] = 1

        def discard(self, n1, n2):
            try:
                del self[n1, n2]
            except KeyError:
                pass


class Directed(Graph):
    
    def outgoing(self, node):
        return {n2: v for (n1, n2), v in self.edges.items() if n1 == node}

    def incoming(self, node):
        return {n1: v for (n1, n2), v in self.edges.items() if n2 == node}

    neighbors = outgoing
    
class Undirected(Graph):

    def neighbors(self, node):
        res = {}
        for (n1, n2), v in self.edges.iteritems():
            if n1 == node:
                res[n2] = v
            elif n2 == node:
                res[n1] = v
        return res

def reversed(g: Directed):
    """Returns a copy of the graph with edge direction reversed."""
    
    return type(g).build((((n2, n1), v) for (n1, n2), v in g.edges.items()),
                         g.nodes)
                             



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module gets information from the Catalogue of Life Annual Checklist
(currently the 2010 version). Browse the checklist at:
http://www.catalogueoflife.org/annual-checklist/2010/

Information about the RESTful web service is at:
http://webservice.catalogueoflife.org/annual-checklist/2010/search/all
"""

import os.path
from urllib.parse import urlencode
from xml.etree import ElementTree

from taxonome import Taxon, Name
from taxonome.config import config
from taxonome.taxa.collection import TaxaResource
from taxonome.taxa.taxonshelf import TaxonShelf
from .utils import urlopen

BASE_URL = "http://www.catalogueoflife.org/annual-checklist/2010/webservice?"

CACHE_FILE = os.path.join(os.path.expanduser(config['cache']['location']), 'CoL_cache')

cache = TaxonShelf(CACHE_FILE)

def _make_taxon(xmldesc, name=None):
    "Make a Taxon object from a suitable XML tree."
    if not name:
        name = _make_name(xmldesc)
    newtaxon = Taxon(name)
    newtaxon.url = xmldesc.find("url").text
    newtaxon.id = xmldesc.find("id").text
    for synonym in xmldesc.findall("synonyms/synonym"):
        synname = _make_name(synonym)
        newtaxon.othernames.add(synname)
    xdistrib = xmldesc.find("distribution")
    distribution = xdistrib.text if (xdistrib is not None) else None
    if distribution:
        newtaxon.distribution = set(distribution.split("; "))
    cache.add(newtaxon)
    return newtaxon

def _make_name(xmldesc):
    "Make a Name object from a suitable XML tree."
    barename = xmldesc.find("name").text
    xauth = xmldesc.find("author")
    auth = xauth.text if (xauth is not None) else ""
    name = Name(barename, auth)
    return name

def _get_xml(name):
    "Get the XML for a given name."
    query = urlencode({'name':name, 'response':'full'})
    return ElementTree.parse(urlopen(BASE_URL+query))

def _get_name_results(name):
    "Find the (name, accepted) name pairs for a given unqualified name."
    results = []
    for result in _get_xml(name).findall("result"):
        status = result.find("name_status").text
        if status == "accepted name":
            accname = _make_name(result)
            results.append((accname, accname))
        
        elif status == "synonym":
            synname = _make_name(result)
            accxml = result.find("accepted_name")
            accname = _make_name(accxml)
            results.append((synname, accname))
            
            # We call this so the taxon gets added to the cache
            _make_taxon(accxml)
        
        else:
            msg = "Unrecognised name status: {0}".format(status)
            raise NotImplementedError(msg) 
            
    return results

def fetch(search):
    """Fetch taxa matching a search term, e.g. "Restionacae". Returns a TaxonSet.
    
    Note that Catalogue of Life only looks one level down, so searching for a
    family will return a collection of genera. There will be minimal information
    about each.
    """
    ts = TaxonSet()
    for xtax in _get_xml(name).findall("result/child_taxa/taxon"):
        ts.add(_make_taxon(xtax))
    return ts

class CoLTaxaResource(TaxaResource):
    def resolve_name(self, key):
        if key in cache:
            return cache.resolve_name(key)
        if isinstance(key, Name):
            return [(n, an) for n, an in _get_name_results(key.plain) \
                                            if n.authority==key.authority]
        return _get_name_results(key)
    
    def get_by_accepted_name(self, name):
        if name in cache:
            return cache.get_by_accepted_name(name)
        tax = _make_taxon(_get_xml(name.plain).find('result'))
        if not name.match_qualified(tax.name):
            raise KeyError("Name did not match with authority.")
        return tax
    
_instance = CoLTaxaResource()

resolve_name = _instance.resolve_name
select = _instance.select
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import shelve

from .collection import TaxonSet

class TaxonShelf(TaxonSet):
    """A persistent version of TaxonSet, storing items using the shelve module.
    
    Initialise using the the name of a database, e.g.:
    mygenus = taxonome.taxon.TaxonShelf("mygenus")
    
    The database consists of two files, with the suffixes "-names" and "-taxa".
    If they do not exist, they will be created.
    
    Modifying taxa on the shelf won't work--instead, make a temporary
    copy of the taxon, then put it back on the shelf:
    temp = myshelf[Name("Lablab purpureus","(L.) Sweet")]
    temp.distribution.add("Somewhere")
    myshelf.add(temp)
    
    N.B. If you change the accepted name of a taxon, you will need to remove
    the old taxon
    """
    def __init__(self, db_name):
        self.names = shelve.open(db_name+"-names")
        self.taxa = shelve.open(db_name+"-taxa")
        
    def sync(self):
        self.names.sync()
        self.taxa.sync()
    
    def __del__(self):
        self.names.close()
        self.taxa.close()
        
    def _add_qname(self, qname, acceptedname=None):
        """Add a qualified name to the index. To add a synonym, specify an
        acceptedname to which it maps. The acceptedname should refer to a taxon
        which is in the set.
        """
        if not acceptedname:
            acceptedname = qname
        if not qname.plain in self.names:
            self.names[qname.plain] = [(qname, acceptedname)]
        else:
            temp = self.names[qname.plain]
            temp.append((qname, acceptedname))
            self.names[qname.plain] = temp
            
    def _del_qname(self, qname):
        """Remove a qualified name from the index. Taxa with no remaining
        names in the index will still take up space, but cannot easily be
        accessed, so you should ensure that at least one name remains for each.
        """
        temp = self.names[qname.plain]
        for namematch in temp:
            if namematch[0] == qname:
                temp.remove(namematch)
                if temp:
                    self.names[qname.plain] = temp
                else:
                    del self.names[qname.plain]
                return namematch[1]
        raise KeyError(qname)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re

from .author import Authority
from taxonome.utils.htmlrepr import htmlise

NoneType = type(None)

class UncertainSpeciesError(ValueError):
    pass

# Used by Name.from_string()
oddauthbits = {"sensu","anon.","auct."}
extranamebits = {"ssp.", "subsp.", "var."}
unknownnamebits = {"spp.", "sp."}
def _isauthbit(txt):
    return txt[0].isupper() or txt.startswith("(") or txt.startswith("&")\
        or txt in oddauthbits

minor_ranks = {'subsp.': 'subspecies', 'var.': 'variety'}

name_re = re.compile(r"""(?P<hybr_g>[×x]\ ?)?
(?P<g>[\w\-]+)
(\ (?P<hybr_sp>×\ ?|x\ )?
 (?P<sp>[\w\-]+)
 (\s+(?P<subsp>[\w\-]+))?
)?""", re.VERBOSE)

subtax_re = re.compile(r"""\b(?P<rank>s(ub)?sp|var)\.? (?P<name>[\w\-]+)""")

class Name(object):
    """Store a species name, optionally qualified with an authority.
    
    When comparing two names, the authority of each will be simplified
    to try to overcome differences in format and punctuation. For example::
    
      a = Name("Lablab purpureus", "(L.) Sweet")
      b = Name("Lablab purpureus","Sweet.")
      c = Name("Lablab purpureus", "Sweet. ex. MadeUp.")
      a == b
      a == c
      b == c    # All True
    """
    hybrid = False   # Default
    _parent = None
    sp = None
    
    def __init__(self, name, authority="", rank=None):
        self.authority = Authority(authority)
        
        if isinstance(name, str):
            namematch = name_re.match(name)
            if namematch is None:
                raise ValueError("Could not parse name", name)
            self.g, self.sp, subsp = namematch.group("g", "sp", "subsp")
            rankguess = 'species' if self.sp else 'genus'
        
            hybr_g, hybr_sp = namematch.group("hybr_g", "hybr_sp")
            if hybr_sp:
                self.hybrid = "species"
            elif hybr_g:
                self.hybrid = "genus"
            
            subnames = []
            if subsp and not subtax_re.search(name):
                # Animal style name with subspecies: Gorilla gorilla gorilla
                rankguess = "subspecies"
                subnames.append(("subsp.", subsp))
            for subtax_match in subtax_re.finditer(name):
                # Plant style name with subspecies: Alloteropsis semialata subsp. semialata
                r, n = subtax_match.group("rank", "name")
                r = r + '.'
                if r == 'ssp.': r = 'subsp.'
                rankguess = minor_ranks[r]
                subnames.append((r, n))
            self.subnames = tuple(subnames)
        else:
            # We were given name parts:
            if len(name) == 1:  # genus
                self.g = name[0]
                rankguess = 'genus'
                self.subnames = ()
            else:
                self.g, self.sp = name[:2]
                rankguess = 'species'
                self.subnames = tuple(tuple(p) for p in name[2:])
                if self.subnames:
                    rankguess = minor_ranks[self.subnames[-1][0]]
        
        self.rank = rank or rankguess
    
    @property
    def plain(self):
        """The unqualified name as a string."""
        parts = [self.g, self.sp] if self.sp else [self.g]
        for sn in self.subnames:
            parts.extend(sn)
        return " ".join(parts)
    
    @property
    def _nameparts(self):
        parts = (self.g, self.sp) if self.sp else (self.g,)
        return parts + self.subnames
    
    @property
    def parent(self):
        """The name one rank up, i.e. species -> genus."""
        if not self._parent:
            self._parent = self._make_parent()
        return self._parent
    
    def _make_parent(self):
        if not self.sp:
            return None     # Can't give a parent for a genus.
        return Name(self._nameparts[:-1])
    
    def __str__(self):
        if self.authority:
            return self.plain + " " + str(self.authority)
        return self.plain
        
    def __repr__(self):
        return "taxonome.Name(\"{0}\", \"{1}\")".format(self.plain, str(self.authority))
    
    _html_template = "<i>{}</i> {}"
    
    def html(self):
        return self._html_template.format(self.plain, str(self.authority))
                
    def __eq__(self, other):
        return (self._nameparts == other._nameparts) \
                and (self.authority == other.authority)
                
    def __ne__(self, other):
        return not self.__eq__(other)
        
    def __hash__(self):
        # ^ is bitwise XOR
        return hash(self._nameparts) ^ hash(self.authority)
    
    @property
    def is_nominal_subsp(self):
        """True if this is a subspecies or variety, and is the nominal form of
        the species, e.g. Vigna unguiculata subsp. unguiculata.
        """
        return self.subnames and (self.subnames[-1][1] == self.sp)
        
    def match(self, other):
        """Compare the name with another qualified name as a Name object, or an
        unqualified name as a string."""
        if isinstance(other, str):
            return self.match_unqualified(other)
        return self == other
        
    def match_unqualified(self, other):
        """Compares the unqualified name with another Name object or a string. E.g.
        
        a = Name("Lablab purpureus", "(L.) Sweet")
        a.match_unqualified("Lablab purpureus")    # True
        """
        if not hasattr(other, "_nameparts"):
            other = Name(other)
        return self._nameparts == other._nameparts
        
    def match_qualified(self, other):
        """Compares the name, including authority, with another Name object or a
        string. If a string is given, it must match the string form of the name
        exactly. With two Name objects, the usual fuzzy logic applies. E.g.
        
        a = Name("Lablab purpureus", "(L.) Sweet")
        a.match_unqualified("Lablab purpureus (L.) Sweet")    # True
        """
        if not hasattr(other, "_nameparts"):
            other = Name(other)
        return self == other
    
    def to_dict(self):
        """Produce a simple dict from this name, suitable for serialising it
        as JSON."""
        d = dict(nameparts=self._nameparts, authority=str(self.authority),)
        if self.hybrid:
            d['hybrid'] = self.hybrid
        return d
    
    @classmethod
    def from_dict(cls, d):
        """Load a Name from a dictionary."""
        nameparts = d['nameparts']
        name = cls(nameparts, d['authority'])
        if 'hybrid' in d:
            name.hybrid = d['hybrid']
        return name
    
    @classmethod
    def from_string(cls, s):
        """Attempts to parse a biological name and authority from a raw string
        of the name, with no other hints to separate them.
        
        An UncertainSpeciesError will be raised if the name is like 'Festuca sp.'
        or 'Festuca spp.'
        """
        namebits = []
        authbits = s.split()
        namebits.append(authbits.pop(0))
        #lowercase first letter --> name
        while authbits and not _isauthbit(authbits[0]):
            namebits.append(authbits.pop(0))
        
        # Handles cases where a species is named with its authority, but the name
        # continues into a subspecies or variety. We discard the species authority.
        parent_authbits = []
        while extranamebits.intersection(authbits):
            parent_authbits = []
            while authbits[0] not in extranamebits:
                parent_authbits.append(authbits.pop(0))
            while authbits and not _isauthbit(authbits[0]):
                namebits.append(authbits.pop(0))
        
        if authbits and authbits[0].startswith("&"):
            raise UncertainSpeciesError("Name not recognised", s)
        if unknownnamebits.intersection(namebits):
            raise UncertainSpeciesError("Invalid sp. or spp. name", s)
        
        name = cls(" ".join(namebits), " ".join(authbits))
        if name.parent and parent_authbits:
            name._parent.authority = Authority(" ".join(parent_authbits))
        return name
    

class Taxon(object):
    url = None
    def __init__(self, name, authority=""):
        if isinstance(name, Name):
            self.name = name
        else:
            self.name = Name(name, authority)
        self.othernames = set()
        self.individuals = []
        self.info = {} #Store other information here
        self.distribution = set() # e.g. For TDWG codes
        
    def __repr__(self):
        return "<Taxon: "+str(self.name)+">"
    
    _html_template = ("<h3>{name}</h3>\n"
                      "{info}\n{individuals}\n{distrib}\n{synonyms}")
    
    def html(self):
        distrib, synonyms, individuals = "", "", ""
        if self.distribution:
            distrib = "<h4>Distribution</h4>{}".format(htmlise(self.distribution))
        if self.othernames:
            synonyms = "<h4>Synonyms</h4>\n{}".format(htmlise([n.html() for n in self.othernames]))
        if self.individuals:
            individuals = "{} individual records".format(len(self.individuals))
        return self._html_template.format(name=self.name.html(), info=htmlise(self.info),
                            distrib=distrib, individuals=individuals, synonyms=synonyms)
        
    def hasname(self, name):
        """Tests if the given name is among the names of this taxon. This
        expects either something like a Name object (with a name attribute) or
        an unqualified name as a string."""
        if isinstance(name, Name):
            if name == self.name:
                return True
            if any(name == othername for othername in self.othernames):
                return True
                
        else:   # Assume the object to compare is like a plain string
            if name == self.name.plain:
                return True
            if any(name == othername.plain for othername in self.othernames):
                return True
                
        return False
    
    def to_dict(self):
        """Produce a simple dict of this taxon's information, suitable for
        serialising as JSON."""
        info = {}
        for k, v in self.info.items():
            if not isinstance(v, (str, int, float, NoneType, bool, dict, list)):
                try:
                    v = str(v)
                except Exception:
                    continue
            info[k] = v
        othernames = [n.to_dict() for n in self.othernames]
        d = dict(name=self.name.to_dict())
        if info:              d['info'] = info
        if othernames:        d['othernames'] = othernames
        if self.url:          d['url'] = self.url
        if self.distribution: d['distribution'] = list(self.distribution)
        if self.individuals:  d['individuals'] = list(self.individuals)
        return d
    
    @classmethod
    def from_dict(cls, d):
        name = Name.from_dict(d['name'])
        t = cls(name, d.get('url'))
        t.info = d.get('info', {})
        t.distribution = set(d.get('distribution', []))
        t.othernames = set(Name.from_dict(n) for n in d.get('othernames', []))
        t.individuals = d.get('individuals', [])
        return t



# -*- coding: utf-8 -*-
from .taxa import Taxon, Name, TaxonSet
from . import config, taxa, regions

__all__ = ['config', 'taxa', 'regions', 'wikipedia', 'col']

#caches = [wikipedia.cache]
#def clean_caches():
#    for cache in caches:
#        cache.clean_outdated()

#if config.getboolean('cache', 'clean-on-load'):
#    clean_caches()
